# -*- coding: utf-8 -*-
"""
Created on Thu Jun 15 15:59:24 2017
@author: Peter Meisrimel
"""
from __future__ import division
from dolfin import *
import numpy as np

## class for timestep controllers and functions to find initial timesteps
class Timestep_controller():
    ## initialization function
    # @param self pointer to self
    # @param tol tolerance used in controller
    # @param t_end end time of problem
    # @param f_min lower limit for change rate of timestep
    # @param f_max upper limit for change rate of timestep
    # @param f_safety safety factor to multiply every timestep with
    # @param get_err function to get error indicator 
    # @param order_low order of lower order method
    # @param num_init_steps number of step repetitions of first timestep to find suitable initial stepsize
    # @param get_y0 pointer to function for providing initial values of problem
    # @param get_eval_err function to do a step and get error estimate, needed for initial step determination
    # @param j_val_func Function to (re)set j value in problem
    # @return None
    def __init__(self, tol = 1, t_end = 0, f_min = 0, f_max = np.inf, f_safety = 1, get_err = None, order_low = 1,
                 num_init_steps = 5, get_y0 = None, get_eval_err = None, j_val_func = None):
        self.tol = tol
        self.t_end = t_end
        self.f_min = f_min
        self.f_max = f_max
        self.f_safety = f_safety
        self.get_err = get_err
        self.phat = order_low
        self.num_init_steps = num_init_steps
        self.get_y0 = get_y0
        self.get_eval_err = get_eval_err
        self.j_val_func = j_val_func
        
        self.t = 0
        self.dt_prev = None
    
    ## Deadbeat controller for classic norm based error estimates
    # @param self pointer to self
    # @param y_new y after evaluation of current timestep
    # @param y_old y from previous timestep
    # @param l_err vector of local error
    # @param dt_val size of previous timestep
    # @return new timestepsize
    def deadbeat_classic(self, y_new, y_old, l_err, dt_val):
        self.t += dt_val
        j_err = abs(np.float64(self.get_err(y_new, y_old, l_err, self.t, dt_val, self.dt_prev if self.dt_prev is not None else dt_val)))
        ind = (self.f_safety * self.tol/j_err)**(1/(self.phat + 1))
        self.dt_prev = dt_val
        return dt_val * max(min(ind, self.f_max), self.f_min)
    
    ## half timesteps until one is found meeting the local error tolance for initial timestep
    # @param self pointer to self
    # @param cont_type type of controller being used
    # @return size of initial timestep
    def dt0_deadbeat(self, cont_type):
        dt_val = 0.5*self.t_end
        while dt_val > 1.e-14:
            y_new, l_err = self.get_eval_err(self.get_y0(), 0, dt_val) # get error est, includes timestep for RK
            err = abs(np.float64(self.get_err(y_new, self.get_y0(), l_err, 0, dt_val, self.dt_prev if self.dt_prev is not None else dt_val)))
            
            if (err != 0.0) and (err < self.tol): # bound fulfilled 
                break
            
            dt_val *= 0.5 # else, bound not fulfilled
        else:
            raise ValueError('initial timestep smaller than 1e.-14')
        self.j_val_func(0)
        return dt_val        
    
    ## take tolerance with appropriate power as initial timestep
    # @param self pointer to self
    # @param cont_type type of controller being used
    # @return size of initial timestep: tol 
    def dt0_tol(self, cont_type):
        self.j_val_func(0)
        return self.tol**(1/(self.phat + 1))
        
    ## take tolerance as initial timestep
    # @param self pointer to self
    # @param cont_type type of controller being used
    # @return size of initial timestep: tol 
    def dt0_tol_tol(self, cont_type):
        self.j_val_func(0)
        return self.tol
        
    ## repeat first timestep "num_init_steps" times and adapt timestep every time
    # @param self pointer to self
    # @param cont_type type of controller being used
    # @return size of initial timestep
    def dt0_repeater(self, cont_type):
        cont = self.get_controller(cont_type)
        dt_val = self.tol
        for i in range(self.num_init_steps):
            y_new, l_err = self.get_eval_err(self.get_y0(), 0, dt_val) # get error est, includes timestep for RK
            dt_val = cont(y_new, self.get_y0(), l_err, dt_val)
        self.dt_prev = 1
        return dt_val        
    
    ## function to return correct controller
    # @param self pointer to self
    # @param cont_type type of controller being used
    # @return pointer to controller
    def get_controller(self, cont_type):
        if cont_type == 0:
            return self.deadbeat_classic
        else:
            raise ValueError('Timestep controller not defined')

    ## function to return \f$ \Delta t_0 \f$ correct controller
    # @param self pointer to self
    # @param cont_type type of controller being used
    # @return pointer to controller
    def get_dt0_controller(self, cont_type):
        if cont_type == 0:
            return self.dt0_tol
        elif cont_type == 1:
            return self.dt0_deadbeat
        elif cont_type == 2:
            return self.dt0_repeater
        elif cont_type == 3:
            return self.dt0_tol_tol
        else:
            raise ValueError('Timestep controller not defined')