# -*- coding: utf-8 -*-
"""
Created on Thu Dec 10 16:09:34 2015
@author: Peter Meisrimel
"""
import os

## get dictionary with default parameters for initialization\n
# for adding new parameters, make sure they are ALL LOWER CASE (!!!)
# @return dictionary with default parameters for input file
def initialize_parameters_default():
    dic = {# Run type
           'run_mode'          : 1,
           'tol_step_start'    : 0,
           'tol_steps'         : 4,
           'tol_steps_dwr'     : None, # number of steps for DWR, if None will use same as tol_steps (in run_mode = 1) , must be <= tol_steps
           'dg'                : 1, # 1 for DG, 0 for RK
           'time_int_type'     : 0,
           'precond'           : 'default', # preconditioner for solving linear systems using FEniCS solve, i.e. 'icc', 'ilu', 'default'
           
           # Problem spcifications
           'problem_type'      : 0,
           'problem_t_end'     : None, # default must be given by problem
           'problem_stiff_para': -1, # stiffness parameter
           'problem_meshsize'  : 32, # parameter to control number of points in space
           'problem_domain'    : 1, # subdomains for space dependent problems
           
           # Parameters for specific problems
           'direction'         : 1, # direction of velocity for heat advection problem
           'fsi_d1'            : 0.1, # Capital D in licentiate thesis/paper
           'fsi_d2'            : 1,
           'fsi_k1'            : 0.1, # lambda in licentiate thesis/paper
           'fsi_k2'            : 1,
           'fsi_tol_fixpoint'  : 1e-4, # tolerance for fixpoint iteration
           'fsi_flux'          : 1, # base flux calculation on which subdomain, 1 for left, 2 for right
           'fsi_u0'            : 5, # initial condition for 2-D FSI problem
           
           # DWR stuff
           'dwr_scheme'        : 0, # 1 = DG0, 2 = CG1, 0 = no DWR
           'dwr_refine_perc'   : 0.8,
           'dwr_start_cells'   : 10, # number of cells in initial grid
           
           # timestep controller
           'dt_cont_type'      : 0, # deadbeat controller
           'dt_cont_f_min'     : 1e-2,
           'dt_cont_f_max'     : 3.,
           'dt_cont_f_safety'  : 1, # commonly used safety factor used for timestep controllers
           'dt_cont_tol'       : 1e-3,
           
           # choice of initial timestep
           'dt_cont_dt0'       : 0, # scheme used for determining the initial timestep
           'num_init_steps'    : 5, # scheme used for determining initial timestep
           
           # error estimation
           'err_est_type'      : 0, # type of error estimator used, e.g. specific areas or components
           
           # functionals used
           'func_type'         : 0,
           
           # number of steps for adjoint solution (run mode 101), will take 'dwr_start_cells' if not set
           'adjoint_steps'     : None,
           
           # printing & plotting options
           'print'             : 1, # save plots as .png
           'legend'            : 'missing legend entry', # legend text
           'path'              : None, # determined by input parameters in main.py
           'saving_path'       : None, # determined by input parameters in main.py
           'filename_self'     : None, # determined by input parameters in main.py
           'save_results'      : 0, # save results to txt file
           'file_ending'       : '.png',
           
           # auxiliarry variables to help creating input files
           'aux1'                 : 0,
           'aux2'                 : 0,
           'aux3'                 : 0,
           'aux4'                 : 0}
    return dic
    
## Takes file reading, removes invalid parameters and overwrites the default values\n
# Base level are the default values, the function will ignore all keys in file reading with no default value (will produce warning)
# and overwrite accepted all keys where values are provided
# @param file_read parameters read from file
# @param para_default reference values to overwrite
# @return None
def initialize_overwrite_default(file_read, para_default):
    rem = [] # keys to be removed from file_read
    for key in file_read.keys():
        if key not in para_default.keys():
            print('Warning : parameter: {} \nis no valid key for initialization, check spelling'.format(key))
            rem.append(key)
    for key in rem:
        file_read.pop(key, None) # remove key from dictionary
    
    for key in para_default.keys():
        if key not in file_read.keys():
            file_read[key] = para_default[key]

## Function to create multiple input files by taking in keys to vary\n
# see examples below on how to conviniently create input files
# @param filename template name of output file
# @param dic dictionary of input data, min. one value needs to be a list
# @param *args keys to be varied, iterates through lists given in dic, succ. filling {}s in filename, lists for parallel list iteration
# @return None, creates lots of files though
def input_file_generator(filename, dic, *args):
    default = initialize_parameters_default()
    if len(args) == 0:
        # check if filename contains subdirectory and create those if needed
        if '/' in filename:
            direct = filename[:filename.rfind('/')]
            if not os.path.exists(direct):
                os.makedirs(direct)
                
        in_file = open(filename, 'w')
        print('creating : ' + filename)
        
        in_file = open(filename, 'w')
        for i in dic.keys():
            if dic[i] != default[i]:
                in_file.write(i + ' = ' + str(dic[i]) + '\n')
    else:
        lista = list(args[0])
        iter_list = dic[lista[0]]
        dic_new = dic.copy()
        for i in range(len(iter_list)):
            for j in lista:
                dic_new[j] = dic[j][i]
            input_file_generator(filename.replace('{}', str(dic[lista[0]][i]), 1), dic_new, *args[1:])

if __name__ == '__main__':
    ############# BASIC VERIFICATION
    input_file_generator('input_files/run_mode_verification/run_mode_-1.txt',  # run_mode = -1 test
                         {'run_mode': -1, 'problem_type': 12, 'problem_t_end': 1,
                         'problem_meshsize': 32, 'func_type': 1, 'dt_cont_tol': 1e-4})
    input_file_generator('input_files/run_mode_verification/run_mode_0.txt',   # run_mode = 0 test
                        {'run_mode': 0, 'problem_t_end': 1, 'problem_stiff_para': -1,
                         'func_type': 1, 'legend': r'$u_1$', 'save_results': 1, 'tol_steps': 7})
    input_file_generator('input_files/run_mode_verification/run_mode_1.txt',   # run_mode = 1 test
                        {'run_mode': 1, 'problem_t_end': 1, 'problem_stiff_para': -1,
                         'func_type': 1, 'legend': r'$u_1$', 'save_results': 1, 'tol_steps': 5, 'tol_steps_dwr': 4, 'dwr_scheme': 2})
    input_file_generator('input_files/run_mode_verification/run_mode_2.txt',   # run_mode = 2 test
                        {'run_mode': 2, 'problem_t_end': 1, 'problem_stiff_para': -1,
                         'func_type': 1, 'legend': 'test', 'save_results': 1, 'tol_steps': 4, 'dwr_scheme': 0})
    input_file_generator('input_files/run_mode_verification/run_mode_10.txt',  # run_mode = 10 test
                        {'run_mode': 10, 'problem_t_end': 1, 'func_type': 1, 'problem_stiff_para': -1,
                         'problem_type': 0, 'tol_steps': 5})
    ## RK
    input_file_generator('input_files/run_mode_verification/RK/run_mode_0.txt',   # run_mode = 0 test RK
                        {'run_mode': 0, 'problem_t_end': 1, 'problem_stiff_para': -1,
                         'func_type': 1, 'legend': r'$u_1$', 'save_results': 1, 'tol_steps': 7, 'dg': 0, 'time_int_type': 21,
                         'problem_type': 0})
    ##  PROBLEMS VERIFICATION
    input_file_generator('input_files/run_mode_verification/toy/run_mode_0.txt',  # run_mode = 0 toy
                        {'run_mode': 0, 'problem_type': 0, 'problem_t_end': 2, 'tol_steps': 4, 'problem_stiff_para': -100,
                         'func_type': 1, 'dg': 1, 'save_results': 1, 'legend': '$u_1$'})
    input_file_generator('input_files/run_mode_verification/ver/run_mode_0.txt', # run_mode = 0 verification problem
                        {'tol_steps': 4, 'problem_type': -1, 'run_mode': 0, 'legend': '$u$', 'problem_t_end': 1,
                         'func_type': 1})
    input_file_generator('input_files/run_mode_verification/ver_time/run_mode_0.txt', # run_mode = 0 verification problem time
                        {'tol_steps': 4, 'problem_type': -2, 'run_mode': 0, 'legend': '$u$', 'problem_t_end': 1,
                         'func_type': 1})
    input_file_generator('input_files/run_mode_verification/heat_src_adv/run_mode_0.txt', # run_mode = 0 verification problem time
                        {'tol_steps': 3, 'problem_type': 12, 'run_mode': 0, 'legend': '$u$', 'problem_t_end': 6,
                         'func_type': 1})
    input_file_generator('input_files/run_mode_verification/FSI_2D/run_mode_0.txt', # run_mode = 0 FSI 2D problem
                        {'tol_steps': 3, 'problem_type': 21, 'run_mode': 0, 'legend': '$u$', 'problem_t_end': 1,
                         'dg': 0, 'problem_meshsize': 20, 'dt_cont_dt0': 0, 'fsi_tol_fixpoint': 1e-5,
                         'fsi_d1': 0.5, 'fsi_d2': 1, 'fsi_k1': 0.3, 'fsi_k2': 1})                         
                         
    ## ADJOINT VERIFICATION
    input_file_generator('input_files/run_mode_verification/adjoints/problem_{}.txt',  # run_mode = 101 test
                        {'run_mode': 101, 'problem_t_end': 2, 'func_type': 1, 'problem_stiff_para': -1,
                         'problem_type': [-2, -1, 0, 1], 'adjoint_steps': 100}, ['problem_type'])
    input_file_generator('input_files/run_mode_verification/adjoints/problem_{}.txt',  # run_mode = 101 test
                        {'run_mode': 101, 'problem_t_end': 6, 'func_type': 1,
                         'problem_type': [12], 'adjoint_steps': 100}, ['problem_type'])
    ############# 0  TOY PROBLEM
    ## RUN MODE   0, basic testing to show convergence
    input_file_generator('input_files/0_basic_testing/toy/k_{}_t_2/func_{}.txt',
                        {'run_mode': 0, 'problem_type': 0, 'problem_t_end': 2, 'tol_steps': 9, 'problem_stiff_para': [-1, -100],
                         'func_type': [1, 2, 10, 21, 22, 31, 32], 'dg': 1, 'save_results': 1, 'file_ending': '.eps',
                         'legend': ['$u_1$','$u_2$','$u_1 u_2$','$u_1^2$','$u_2^2$','$u_1^3$','$u_2^3$']},
                         ['problem_stiff_para'], ['func_type', 'legend'])
    ## RUN_MODE   1, tolerance comparison of various controllers and DWR
    input_file_generator('input_files/1_tolerance_DWR/toy/k_{}/func_{}.txt',
                        {'run_mode': 1, 'problem_type': 0, 'problem_t_end': 2, 'tol_steps': [8,7], 'file_ending': '.eps',
                         'func_type': [1, 2, 10, 21, 22, 31, 32], 'problem_stiff_para': [-1, -100],
                         'save_results': 1, 'dwr_scheme': 2}, ['problem_stiff_para', 'tol_steps'], ['func_type'])
    ############# 12 HEAT SOURCE PROBLEM ADVECTION
    ## RUN_MODE  -1, reference solution
    input_file_generator('input_files/-1_reference/heat_src_advection/dx_32_t_6/tol_{}.txt',
                        {'run_mode': -1, 'problem_type': 12, 'problem_t_end': 6,
                         'problem_meshsize': 32, 'func_type': 1, 'dt_cont_tol': [10**(-i) for i in range(5,9)],
                         'problem_domain': 1, 'save_results': 1}, ['dt_cont_tol'])
                         
    input_file_generator('input_files/-1_reference/heat_src_advection_reverse/dx_32_t_3/tol_{}.txt',
                        {'run_mode': -1, 'problem_type': 12, 'problem_t_end': 3,
                         'problem_meshsize': 32, 'func_type': 1, 'dt_cont_tol': [10**(-i) for i in range(5,12)],
                         'problem_domain': 1, 'save_results': 1, 'direction': -1}, ['dt_cont_tol'])
    ## RUN MODE   0, basic testing to show convergence
    input_file_generator('input_files/0_basic_testing/heat_src_advection/func_{}.txt',
                        {'tol_steps': 10, 'problem_type': 12, 'run_mode': 0, 'legend': '$u$', 'problem_t_end': 6,
                         'func_type': [1], 'problem_meshsize': 32, 'problem_domain': 1},
                         ['func_type'])
    input_file_generator('input_files/0_basic_testing/heat_src_advection_reverse/func_{}.txt',
                        {'tol_steps': 5, 'problem_type': 12, 'run_mode': 0, 'legend': '$u$', 'problem_t_end': 3,
                         'func_type': [1], 'problem_meshsize': 32, 'problem_domain': 1, 'direction': -1},
                         ['func_type'])
    ## RUN_MODE   1, tolerance comparison of various controllers and DWR
    input_file_generator('input_files/1_tolerance_DWR/heat_src_advection/func_{}.txt',
                        {'problem_t_end': 6, 'tol_steps': 5, 'problem_type': 12, 'tol_steps_dwr': 5,
                         'func_type': [1], 'problem_meshsize': 32, 'problem_domain': 1,
                         'dwr_scheme': 2, 'file_ending': '.eps', 'save_results': 1}, ['func_type'])
    input_file_generator('input_files/1_tolerance_DWR/heat_src_advection_reverse/func_{}.txt',
                        {'problem_t_end': 3, 'tol_steps': 7, 'problem_type': 12, 'tol_steps_dwr': 7,
                         'func_type': [1], 'problem_meshsize': 32, 'problem_domain': 1, 'direction': -1,
                         'dwr_scheme': 2, 'file_ending': '.eps', 'save_results': 1}, ['func_type'])
    ## RUN_MODE   2, tolerance comparison of defined input files
    input_file_generator('input_files/2_tolerances/heat_src_advection/forward/func_1/est_{}.txt',
                         {'run_mode': 2, 'tol_steps': [7, 7], 'problem_type': 12, 'func_type': 1, 'problem_meshsize': 32,
                          'problem_t_end': 6, 'direction': 1, 'problem_domain': 1, 'save_results': 1, 
                          'err_est_type': [0,1], 'legend': ['Classic', 'Goal Or.'], 'file_ending': '.eps'},
                         ['err_est_type', 'tol_steps', 'legend'])
    input_file_generator('input_files/2_tolerances/heat_src_advection/backward/func_1/est_{}.txt',
                         {'run_mode': 2, 'tol_steps': [10, 9], 'problem_type': 12, 'func_type': 1, 'problem_meshsize': 32,
                          'problem_t_end': 3, 'direction': -1, 'problem_domain': 1, 'save_results': 1, 
                          'err_est_type': [0,1], 'legend': ['Classic', 'Goal Or.'], 'file_ending': '.eps'},
                         ['err_est_type', 'tol_steps', 'legend'])
    ############# 21 2 D THERMAL FSI PROBLEM
    ## RUN_MODE  -1, reference solution
    input_file_generator('input_files/-1_reference/FSI_2D/u0_{}/ref_tol_{}.txt',
                        {'problem_type': 21, 'run_mode': -1, 'problem_t_end': 1, 'dt_cont_tol': [10**(-i) for i in range(3, 8)],
                         'dg': 0, 'problem_meshsize': 20, 'dt_cont_dt0': 3, 'fsi_tol_fixpoint': 1e-10,
                         'fsi_d1': 0.5, 'fsi_d2': 1, 'fsi_k1': 0.3, 'fsi_k2': 1, 'fsi_u0': [1,2,3,4,5]},
                         ['fsi_u0'], ['dt_cont_tol'])
    ## RUN_MODE   2, tolerance comparison of defined input files
    input_file_generator('input_files/2_tolerances/FSI_2D/u0_{}/est_{}.txt',
                        {'tol_steps': 6, 'problem_type': 21, 'run_mode': 2, 'problem_t_end': 1, 'fsi_u0': [2, 3, 4, 5],
                         'legend': ['Classic', 'Goal Or.'], 'save_results': 1,
                         'dg': 0, 'problem_meshsize': 20, 'dt_cont_dt0': 3, 'fsi_tol_fixpoint': 1e-10,
                         'err_est_type': [0,1], 'fsi_d1': 0.5, 'fsi_d2': 1, 'fsi_k1': 0.3, 'fsi_k2': 1, 'file_ending': '.eps'},
                         ['fsi_u0'], ['err_est_type', 'legend'])
