# -*- coding: utf-8 -*-
"""
Created on Thu Aug  3 10:52:12 2017

@author: Azahar Monge, some editing and animation by Peter Meisrimel
"""

from __future__ import division
import numpy as np
import pylab as pl
import scipy.sparse as sp
from scipy.sparse.linalg import spsolve

import locale
_encoding = locale.getpreferredencoding(True)

# FUNCTIONS DEFINITION
class Problem_FSI_2D():
    a = 1 - (1/2)*np.sqrt(2)
    ahat = 2 - (5/4)*np.sqrt(2)
    def __init__(self, n = 10, t_end = 0.5, D1 = 0, D2 = 0, K1 = 0, K2 = 0, fixpoint_tol = None):
        self.n = n
        self.dx = 1/(self.n+1)
        self.t_end = t_end
        self.K1, self.K2 = K1, K2
        self.d1, self.d2 = self.K1/D1, self.K2/D2
        self.fixpoint_tol = fixpoint_tol
        
        self.A1, self.A2, self.M1, self.M2, self.A1g, self.A2g, self.Ag1, self.Ag2, \
        self.Agg1, self.Agg2, self.M1g, self.M2g, self.Mg1, self.Mg2, self.Mgg1, self.Mgg2 = self.compute_matrices()
        
        ## B Matrix for Neumann consists of 2 parts, one depedent, the other one independent of the timestep
        # python requires csr/csc format for lin equation solving, which are horrible for assigning values
        # pythons lil format is good for assigning values, but constant format conversion is costly as well
        # Thus the matrix is split into parts according to time dependence
        # change in timesteps are dealt with via multiplying the variable part by factors
        self.B_fixed = sp.lil_matrix((n**2+n,n**2+n))
        self.B_fixed[0:n**2,0:n**2] = self.A2
        self.B_fixed[n**2:n**2+n,0:n**2] = self.Ag2
        self.B_fixed[0:n**2,n**2:n**2+n] = self.A2g[:,0:n]
        self.B_fixed[n**2:n**2+n,n**2:n**2+n] = self.Agg2
        
        self.B_var = sp.lil_matrix((n**2+n,n**2+n))
        self.B_var[0:n**2,0:n**2] = self.M2
        self.B_var[n**2:n**2+n,0:n**2] = self.Mg2
        self.B_var[0:n**2,n**2:n**2+n] = self.M2g[:,0:n]
        self.B_var[n**2:n**2+n,n**2:n**2+n] = self.Mgg2
        
        self.B_var = self.B_var.tocsr()
        self.B_fixed = self.B_fixed.tocsr()
        
        ## Matrix used in lhs flux calculation
        self.E1 = sp.spdiags(np.ones(n**2), n**2 - n, n, n**2, format = 'csr')
        ## Matrix used in rhs flux calculation
        self.E2 = sp.spdiags(np.ones(n**2), 0, n, n**2, format = 'csr')
        
    def compute_matrices(self):
        n, dx = self.n, self.dx
        A1g = sp.spdiags(-(self.K1/(dx**2))*np.ones(n**2), -n**2 + n, n**2, n, format='csr')
        A2g = sp.spdiags(-(self.K2/(dx**2))*np.ones(n**2), 0        , n**2, n, format='csr')
        
        Ag1 = sp.spdiags(-(self.K1/(dx**2))*np.ones(n**2), n**2 - n , n, n**2, format = 'csr')
        Ag2 = sp.spdiags(-(self.K2/(dx**2))*np.ones(n**2), 0        , n, n**2, format = 'csr')
    
        Agg1 = sp.spdiags([2*self.K1/(dx**2)*np.ones(n)] + 2*[-0.5 * self.K1/(dx**2)*np.ones(n)], [0, 1, -1], n, n, format = 'csr')
        Agg2 = sp.spdiags([2*self.K2/(dx**2)*np.ones(n)] + 2*[-0.5 * self.K2/(dx**2)*np.ones(n)], [0, 1, -1], n, n, format = 'csr')

        M1g = sp.spdiags([-self.d1/12*np.ones(n**2), self.d1/4 *np.ones(n**2)], [-n**2 + n, -n**2 + n - 1], n**2, n, format='csr')
        M2g = sp.spdiags([-self.d2/12*np.ones(n**2), self.d2/4 *np.ones(n**2)], [0, 1], n**2, n, format='csr')
        
        Mg1 = sp.spdiags([-self.d1/12*np.ones(n**2), [0]*(n**2 - n) + [self.d1/4]*n], [n**2 - n, n**2 - n - 1] , n, n**2, format='csr')
        Mg2 = sp.spdiags([-self.d2/12*np.ones(n), self.d2/4 *np.ones(n)], [0, 1], n, n**2, format='csr')
              
        Mgg1 = sp.spdiags([5*self.d1/12*np.ones(n)] + 2*[-self.d1/24*np.ones(n)], [0, 1, -1], n, n, format = 'csr')
        Mgg2 = sp.spdiags([5*self.d2/12*np.ones(n)] + 2*[-self.d2/24*np.ones(n)], [0, 1, -1], n, n, format = 'csr')

        B = sp.spdiags(4*np.ones(n),  0, n, n, format = 'csr') - \
            sp.spdiags(  np.ones((2,n)),  [-1, 1], n, n, format = 'csr') 
        A1 = sp.kron(sp.spdiags(np.ones(n), 0, n, n, format = 'csr'), B) + \
             sp.kron(sp.spdiags(np.ones((2,n)), [-1, 1], n, n, format = 'csr'), sp.spdiags(-np.ones(n), 0, n, n, format = 'csr'))
        A2 = self.K2/(dx**2)*A1
        A1 = self.K1/(dx**2)*A1
        
        N = sp.spdiags([5/6*np.ones(n)] + 2*[-1/12*np.ones(n)], [0, 1, -1], n, n, format = 'csr')
        N1 = sp.spdiags([-1/12*np.ones(n), 1/4*np.ones(n)], [0, -1], n, n, format = 'csr')
        N2 = sp.spdiags([-1/12*np.ones(n), 1/4*np.ones(n)], [0, 1], n, n, format = 'csr')
        M1 = sp.kron(sp.spdiags(np.ones(n), 0, n, n, format = 'csr'), N) + \
             sp.kron(sp.spdiags(np.ones(n), -1, n, n, format = 'csr'), N1) + \
             sp.kron(sp.spdiags(np.ones(n), 1, n, n, format = 'csr'), N2)
        M2 = self.d2*M1
        M1 = self.d1*M1
        return ( A1, A2, M1, M2, A1g, A2g, Ag1, Ag2,
                Agg1, Agg2, M1g, M2g, Mg1, Mg2, Mgg1, Mgg2)
    
    def compute_flux(self, dt, u0, u, ug_0, ug_old):
        flux = -((1/dt)*self.Mg1 + self.Ag1).dot(u) - ((1/dt)*self.Mgg1 + self.Agg1).dot(ug_old) + \
               (1/dt)*self.Mg1.dot(u0) + (1/dt)*self.Mgg1.dot(ug_0)
        return flux
    
    def solve_dirichlet(self,dt ,u0 ,ug_0 ,ug_old):
        u = spsolve(((1/dt)*self.M1+self.A1),(-((1/dt)*self.M1g+self.A1g).dot(ug_old) + \
            (1/dt)*self.M1.dot(u0) + (1/dt)*self.M1g.dot(ug_0)))
        flux = self.compute_flux(dt, u0, u, ug_0, ug_old)
        return u, flux
        
    def solve_neumann(self, dt, u0, ug_0, flux):
        n = self.n
        
        b = np.zeros(n**2+n)
        b[0:n**2] = (1/dt)*self.M2.dot(u0) + (1/dt)*self.M2g.dot(ug_0)
        b[n**2:n**2+n] = flux + (1/dt)*self.Mg2.dot(u0) + (1/dt)*self.Mgg2.dot(ug_0)
        
        # see __init__ for implementation of B Matrix
        u = spsolve((1/dt)*self.B_var + self.B_fixed, b)

        u_new = u[0:n**2]
        ug_new = u[n**2:]
        self.dt_old = dt
        return u_new, ug_new
    
    def compute_flux1(self, u):
        n = self.n
        vecflux = 2*u[n**2 + 1:n**2+n] - self.E1.dot(u[0:n**2])[0:n-1] - u[n**2:n**2+n-1]
        return -self.K1*(sum(vecflux) + 2*u[n**2])
    
    def compute_flux2(self, u):
        n = self.n
        vecflux = (-2*u[n**2:n**2+n-1] + self.E2.dot(u[n**2+n:2*n**2+n])[1:n] + u[n**2+1:n**2+n])
        return self.K2*(sum(vecflux) -2*u[n**2:n**2+n][n-1])
        
    def compute_flux1_dense(self, u):
        n = self.n
        ug = u[n*n:n*n+n]
        uaux = np.zeros(n)
        ugaux = np.zeros(n)
        uaux[1:n] = u[n**2 - n:n**2-1]
        ugaux[1:n] = ug[0:n-1]
        vecflux = 2*ug - uaux - ugaux
        return -self.K1*sum(vecflux)
    
    def compute_flux2_dense(self, u):
        n = self.n
        ug= u[n*n:n*n+n]
        uaux = np.zeros(n)
        ugaux = np.zeros(n)
        uaux[0:n-1] = u[n**2+n+1:n**2+2*n]
        ugaux[0:n-1] = ug[1:n]
        vecflux = -2*ug + uaux + ugaux
        return self.K2*sum(vecflux)
        
    def SDIRK2_step(self, dt, uold):
        n = self.n
        u1_start = uold[0:n**2]
        ug_start = uold[n**2:n**2+n]
        u2_start = uold[n**2+n:2*n**2+n]
        
        # calculating first stage
        error = np.inf
        ug_old = ug_start
        while error > self.fixpoint_tol:
            u1_mid, flux_stage1 = self.solve_dirichlet(self.a*dt, u1_start, ug_start, ug_old)
            u2_mid, ug_mid = self.solve_neumann(self.a*dt, u2_start, ug_start, flux_stage1)
            error = np.linalg.norm(ug_old - ug_mid, 2)
            ug_old = ug_mid
        
        # first stage solution
        k1_u1 = (1/(self.a*dt))*(u1_mid-u1_start)
        k1_u2 = (1/(self.a*dt))*(u2_mid-u2_start)
        k1_ug = (1/(self.a*dt))*(ug_mid-ug_start)
        

        u1_start_stage2 = u1_start + dt*(1-self.a)*k1_u1
        u2_start_stage2 = u2_start + dt*(1-self.a)*k1_u2
        ug_start_stage2 = ug_start + dt*(1-self.a)*k1_ug        
        # calculating first stage
        error = np.inf
        ug_old = ug_start_stage2
        while error > self.fixpoint_tol:
            u1, flux_stage2 = self.solve_dirichlet(self.a*dt, u1_start_stage2, ug_start_stage2, ug_old)
            u2, ug_new = self.solve_neumann(self.a*dt, u2_start_stage2, ug_start_stage2, flux_stage2)
            error = np.linalg.norm(ug_old-ug_new, 2)
            ug_old = ug_new
        
        # second stage
        k2_u1 = (1/(self.a*dt))*(u1-u1_start_stage2)
        k2_u2 = (1/(self.a*dt))*(u2-u2_start_stage2)
        k2_ug = (1/(self.a*dt))*(ug_new-ug_start_stage2)
    
        # solution
        u = np.zeros(2*n**2+n)
        u[0:n**2]          = u1_start + dt*((1-self.a)*k1_u1+self.a*k2_u1)
        u[n**2:n**2+n]     = ug_start + dt*((1-self.a)*k1_ug+self.a*k2_ug)
        u[n**2+n:2*n**2+n] = u2_start + dt*((1-self.a)*k1_u2+self.a*k2_u2)
        
        # local error estimate
        localu = np.zeros(2*n**2+n)
        localu[0:n**2]          = dt*((self.ahat - self.a)*k1_u1 + (self.a - self.ahat)*k2_u1)
        localu[n**2:n**2+n]     = dt*((self.ahat - self.a)*k1_ug + (self.a - self.ahat)*k2_ug)
        localu[n**2+n:2*n**2+n] = dt*((self.ahat - self.a)*k1_u2 + (self.a - self.ahat)*k2_u2)
        
        return u, localu
    
    def DN_SDIRK2_Algorithm(self, N, filename):
        # filename for video file, automatically adds .mp4
        from matplotlib import animation
        
        n = self.n
        dt = self.t_end/N
        
        
        ### INITIAL CONDITIONS
        def initial_condition(x,y):
            #return abs(200*np.sin(np.pi*y**2)*np.sin((np.pi/2)*(x**2)))
            if x >= 1:
                return 200
            return 0
    
        x1 = np.linspace(0,1,n+2)
        xplot = np.linspace(0,2,2*n+3)
    
        U_0 = np.zeros((2*n+1,n))
        for i in range(2*n+1):
            for j in range(n):
                U_0[i,j] = initial_condition(xplot[i+1],x1[j+1])
        u_start = U_0.reshape(n*(2*n+1))
        
        
        class updating:
            def __init__(self, u_start, SDIRK2_step):
                self.u_start = u_start
                self.SDIRK2_step = SDIRK2_step
                self.fig, ax = pl.subplots(figsize = (10, 5))
                ax.set_xlim([0, 2*n+1])
                ax.set_ylim([0, n])
                self.line = ax.pcolor(u_start.reshape(2*n + 1, n).T)
            def init_anim(self):
                self.line.set_array((self.u_start.reshape(2*n + 1, n).T).ravel())
                return self.line
            def animate(self, i):
                self.line.set_array((self.u_start.reshape(2*n + 1, n).T).ravel())
                u,_ = self.SDIRK2_step(dt, self.u_start)
                self.u_start = u
                return self.line
        
        runtime = 5 # sec
        up = updating(u_start, self.SDIRK2_step)
        anim = animation.FuncAnimation(up.fig, up.animate, init_func=up.init_anim,
                               frames=N, interval=runtime/N*1e3/self.t_end, blit=False, repeat = True)
        anim.save(filename + '.mp4', fps=self.t_end*N/runtime, extra_args=['-vcodec', 'libx264'])
        
N = 100 # timesteps
Prob = Problem_FSI_2D(n = 50, t_end = 0.5, D1 = 0.5, D2 = 1, K1 = 0.3, K2 = 1, fixpoint_tol = 1e-6)
Prob.DN_SDIRK2_Algorithm(N, 'one_side_u0')