# -*- coding: utf-8 -*-
"""
Created on Thu Dec 10 16:09:34 2015
@author: Peter Meisrimel
"""
from __future__ import division
from dolfin import *
import numpy as np
set_log_level(WARNING) # to suppress massive ammounts of output

## Head class for defining Problems \n
# contains methods for error estimation and DWR-grid refinement
class Problem():
    ## Constant to use for theta method
    theta = Constant(0.5)
    ## lower order of CN method = order of implicit euler scheme
    order_low = 1
    ## Marker if problem has a time dependant term to take care of in time-integration
    # to be set as True in respective constructors if needed
    time_dep = False
    ## functional value
    j_val = 0.
    ## timestep, needed for variational formulations
    dt = Constant(0)
    ## j evaluation at previous point
    j_old = 0
    ## j evaluation at previous point
    j_new = 0
    
    ## Initialize basics
    # @param self pointer to itself
    # @param DG Set to 1 to use DG, 0 for RK. Note: Not all problem support RK
    # @param dwr_scheme which DWR scheme to use
    # @return None
    def __init__(self, DG = 1, dwr_scheme = 0):
        ## Sets to run DG or RK
        self.DG = DG
        ## DWR scheme to be used
        self.dwr_scheme = dwr_scheme
    
    ## function to reset initial values, needed for finding suitable initial stepsizes
    # @param self pointer to self
    # @param DWR set to True to return pointer to y_old and resetting initial values beforehand (needed for annotation purposes)
    # @param adj set to Ture to get initial value of the adjoint problem
    # @return initial value is according format
    def get_initial_value(self, DWR = False, adj = False):
        if DWR: # needed in the DWR case for annotation purposes
            self.initialize_vectors()
            if adj:
                return self.z_old
            return self.y_old
        if self.DG:
            return self.y0.copy(deepcopy = True)
        else:
            return np.copy(self.y0)
    
    ## function to get the calculated value of J(u)
    # @param self pointer to self
    # @param val if set, it will replace j_val with the given value
    # @return value of J(u)
    def j_val_func(self, val = None):
        if val is not None:
            self.j_val = val
            if val == 0 and not self.dwr_scheme:
                self.initialize_vectors()
        return self.j_val
    
    ## update j value and corresponding stored values used for quadrature estimation
    # @param self pointer to self
    # @param dt_val size of current timestep
    # @param y_new solution at new point
    # @return None, class attributes are being updated
    def j_update(self, dt_val, y_new):
        self.j_old = self.j_new
        self.j_new = self.j(y_new, val = True)
        self.j_val += 0.5 * (self.j_new + self.j_old)*dt_val
        
    ## Initialize run parameters
    # @param self pointer to itself
    # @param precond which preconditioner to use for solving linear systems
    # @return None
    def initialize_vectors(self, precond = 'default'):
        if self.DG:
            self.y0 = interpolate(self.y0_expr, self. V)
            ## initialize space for loval error estimate
            self.y_new = Function(self.y0)
            self.y_old = Function(self.y0)
            self.y_lo  = Function(self.y0)
            self.l_err = Function(self.V)
            self.j_new = self.j(self.y_new, val = True)
            self.j_old = 0
            if self.dwr_scheme > 0:
                self.z_old = interpolate(self.z0_expr, self.V)
        else: # RK
            self.l_err = np.copy(self.y0)
            self.y_new = np.copy(self.y0)
            self.y_old = np.copy(self.y0)
            self.y_lo = np.copy(self.y0)
            self.j_new = self.j(self.y_new, val = True)
            self.j_old = 0
    
    ## function to call for evaluating the new step
    # @param self pointer to self
    # @param y_old y before evaluation of current timestep
    # @param prev_time current time
    # @param dt_val size of current timestep
    # @param no_j_update set True to supress update of j value
    # @return None
    def get_eval(self, y_old, prev_time, dt_val, no_j_update = False):
        if self.DG:
            self.y_old.assign(y_old)
            self.dt.assign(dt_val)
            if self.time_dep:  self.f1.t = prev_time; self.f2.t = prev_time + np.float64(dt_val)
            if self.dwr_scheme:
                self.solver_DWR.solve()
            else:
                self.solver.solve()
            if not no_j_update:
                self.j_update(np.float64(dt_val), self.y_new)
            return self.y_new
        else:
            y_new, _ = self.do_step(y_old, prev_time, dt_val)
            if not no_j_update:
                self.j_update(dt_val, y_new)
            return y_new
            
    ## function to call for evaluating new step including error estimate
    # @param self pointer to self
    # @param y_old y before evaluation of current timestep
    # @param prev_time time before timestep
    # @param dt_val size of current timestep
    # @param no_j_update set True to supress update of j value
    # @return None
    def get_eval_err(self, y_old, prev_time, dt_val, no_j_update = False):
        if self.DG:
            self.y_old.assign(y_old)
            self.dt.assign(dt_val if type(dt_val) != tuple else dt_val[0])
            if self.time_dep:  self.f1.t = prev_time; self.f2.t = prev_time + np.float64(dt_val)
            self.solver.solve()
            if not no_j_update:
                self.j_update(np.float64(dt_val), self.y_new)
            self.solver_err.solve()
            self.l_err.assign(self.y_new - self.y_lo)
            #print('ESTIMATED ERROR: {}'.format(assemble(self.l_err*dx)))
            return self.y_new, self.l_err
        else:
            # time_int needs to be set outside
            y_new, l_err = self.do_step(y_old, prev_time, dt_val)
            if not no_j_update:
                self.j_update(dt_val, y_new)
            return y_new, l_err
    
    ## Solver for DWR
    def get_eval_DWR(self, old_val, prev_time, dt_val, fwd = True):
        self.dt.assign(dt_val)
        if fwd:
            self.y_old.assign(old_val)
            if self.time_dep:  self.f1.t = prev_time; self.f2.t = prev_time + np.float64(dt_val)
            self.solver_DWR_fwd.solve()
            self.j_update(np.float64(dt_val), self.y_new)
            return self.y_new
        else: # adjoint solve, no f based source term
            self.z_old.assign(old_val)
            self.solver_DWR_adj.solve()
            return self.z_new
            
    ## Here: dummy function to prevent error
    # @param self .
    # @param t .
    # @param y_loc .
    # @return None
    def solution_point(self, t = None, y_loc = None):
        pass
            
## header class for general scalar based ODEs
class Problem_scalar(Problem):
    ## function to get error estimation function, e.g. norm or goal oriented
    # @param self pointer to self
    # @param err_est_type type of error estimator
    # @return error estimator function
    def get_j_err_func(self, err_est_type):
        if self.DG:
            if err_est_type == 0:
                def get_j_err(y_new, y_old, l_err, t, dt_val, dt_prev):
                    return np.sqrt(np.dot(l_err.vector().array(), l_err.vector().array()))
            elif err_est_type == 1:
                def get_j_err(y_new, y_old, l_err, t, dt_val, dt_prev):
                    return abs(self.j_new - self.j(y_new.vector().array() - l_err.vector().array(), force = True))
        else: # RK
            if err_est_type == 0:
                def get_j_err(y_new, y_old, l_err, t, dt_val, dt_prev):
                    return np.sqrt(np.dot(l_err, l_err))
            elif err_est_type == 1:
                    def get_j_err(y_new, y_old, l_err, t, dt_val, dt_prev):
                        return abs(self.j_new - self.j(y_new - l_err))
        return get_j_err
    
## header class for general 2-D problems using fenics
class Problem_2D(Problem):
    ## function to get error estimation function, e.g. norm or goal oriented
    # @param self pointer to self
    # @param err_est_type type of error estimator
    # @return error estimator function
    def get_j_err_func(self, err_est_type):
        if err_est_type == 0:
            def get_j_err(y_new, y_old, l_err, t, dt_val, dt_prev):
                return np.sqrt(abs(assemble(inner(l_err, l_err)*dx)))
        elif err_est_type == 1:
                def get_j_err(y_new, y_old, l_err, t, dt_val, dt_prev):
                    return abs(self.j_new - self.j(y_new - l_err, val = True))
        return get_j_err