# -*- coding: utf-8 -*-
"""
Created on Tue Aug 8 15:48:31 2017
@author: Azahar Monge, Peter Meisrimel
"""
from __future__ import division
from problem import Problem
import numpy as np
import scipy.sparse as sp

## Class for fluid structure interaction problem for 2 coupled linear heat equations
class Problem_FSI(Problem):
    ## Initialization basics
    # @param self pointer to itself
    # @param gridsize gridsize for spatial discretization
    # @param t_end end time of problem
    # @param D1 heat equation related parameter, same in licentiate thesis/paper
    # @param D2 heat equation related parameter
    # @param K1 heat equation related parameter, lambda in licentiate thesis/paper
    # @param K2 heat equation related parameter
    # @param flux which subdomain to base flux calculation on
    # @param DG ignore here
    # @param u0 initial condition
    # @return None
    def __init__(self, gridsize = 50, t_end = None, D1=0.1, D2=1, K1=0.1, K2=1, flux = 1, DG = 0, u0 = 0):
        Problem.__init__(self, DG = DG)
        self.n = gridsize
        self.dx = 1/(self.n+1)
        self.t_end = t_end
        self.K1, self.K2 = K1, K2
        self.D1, self.D2 = D1, D2
        self.d1 = self.K1/self.D1 # alpha in licentiate thesis/paper
        self.d2 = self.K2/self.D2
        self.u0 = u0
        
        if t_end is None: self.t_end = 0.5
        else:             self.t_end = t_end
        self.func = 1
        
        self.gridsize = gridsize
        
        if flux == 1:    self.flux = self.compute_flux1
        elif flux == 2:  self.flux = self.compute_flux2
        else:            raise ValueError('invalid flux selected')
        
    ## Density function
    # @param self pointer to itself
    # @param z Vector to evaluate
    # @param val Set to True get return as float
    # @param force Ignore here
    # @return Result of evaluating z with density function
    def j(self, z, val = False, force = False):
        return self.flux(z)
    
    ## function to get error estimation function, e.g. norm or goal oriented
    # @param self pointer to self
    # @param err_est_type type of error estimator
    # @return error estimator function
    def get_j_err_func(self, err_est_type):
        if err_est_type == 0:
            def get_j_err(y_new, y_old, l_err, t, dt_val, dt_prev):
                return np.sqrt(np.dot(l_err, l_err))
        elif err_est_type == 1:
            def get_j_err(y_new, y_old, l_err, t, dt_val, dt_prev):
                return abs(self.j_new - self.j(y_new - l_err, force = True))
        return get_j_err

## Class for fluid structure interaction problem for 2 coupled linear heat equations in 1 dimension
# \n problem_type = 20
class Problem_FSI_1D(Problem_FSI):
    ## Initialization basics
    # @param self pointer to itself
    # @param gridsize gridsize for spatial discretization
    # @param t_end end time of problem
    # @param D1 heat equation related parameter
    # @param D2 heat equation related parameter
    # @param K1 heat equation related parameter
    # @param K2 heat equation related parameter
    # @param flux which subdomain to base flux calculation on
    # @param DG ignore here
    # @param u0 initial condition
    # @return None
    def __init__(self, gridsize = 50, t_end = None, D1 = 0.1, D2 = 1, K1 = 0.1, K2 = 1, flux = 1, DG = 0, u0 = 0):
        Problem_FSI.__init__(self, gridsize = gridsize, t_end = t_end, D1 = D1, D2 = D2, K1 = K1, K2 = K2, flux = flux,
                             DG = DG, u0 = u0)
        
        xplot = np.linspace(0, 2, 2 * self.n + 3)
        u0 = -xplot[1:-1]**2 + 2*xplot[1:-1]
        
        self.y0 = np.zeros((2*self.n+1,1))
        for i in range(2*self.n+1):
            self.y0[i,0] = u0[i]
        self.y0 = self.y0.reshape((len(self.y0), 1))
    
    ## function to compute FE Discretization matrices
    # @param self pointer to self
    # @return self.n and tons of matrices
    def compute_matrices(self): 
        A1g=np.zeros((self.n,1))
        A2g=np.zeros((self.n,1))
        A1g[self.n-1,0] = -(self.K1/(self.dx**2))
        A2g[0,0] = -(self.K2/(self.dx**2))
        
        Ag1 = np.zeros((1,self.n))
        Ag2 = np.zeros((1,self.n))
        Ag1[0,self.n-1] = -self.K1/(self.dx**2)
        Ag2[0,0] = -self.K2/(self.dx**2)
    
        Agg1 = self.K1/(self.dx**2)
        Agg2 = self.K2/(self.dx**2)
    
        M1g = np.zeros((self.n,1))
        M2g = np.zeros((self.n,1))
        M1g[self.n-1,0] = self.d1/6
        M2g[0,0] = self.d2/6
    
        Mg1 = np.zeros((1,self.n))
        Mg2 = np.zeros((1,self.n))
        Mg1[0,self.n-1] = self.d1/6
        Mg2[0,0] = self.d2/6
    
        Mgg1 = self.d1/3
        Mgg2 = self.d2/3
        
        A1 = np.diag(self.n*[2*self.K1/(self.dx**2)]) + np.diag((self.n-1)*[-(self.K1/(self.dx**2))],-1) + np.diag((self.n-1)*[-(self.K1/(self.dx**2))],1)
        A2 = np.diag(self.n*[2*self.K2/(self.dx**2)]) + np.diag((self.n-1)*[-(self.K2/(self.dx**2))],-1) + np.diag((self.n-1)*[-(self.K2/(self.dx**2))],1)
        
        M1 = np.diag(self.n*[4*self.d1/6]) + np.diag((self.n-1)*[self.d1/6],-1) + np.diag((self.n-1)*[self.d1/6],1)
        M2 = np.diag(self.n*[4*self.d2/6]) + np.diag((self.n-1)*[self.d2/6],-1) + np.diag((self.n-1)*[self.d2/6],1)
    
        return self.n, A1, A2, M1, M2, A1g, A2g, Ag1, Ag2, Agg1, Agg2, M1g, M2g, Mg1, Mg2, Mgg1, Mgg2
    
    ## computation of flux over interface based on evaluation in first region
    # @param self pointer to self
    # @param u current state
    # @return flux over boundary calculated from left side
    def compute_flux1(self, u):
        return -(self.K1/self.dx)*(u[self.n]-u[self.n-1])

    ## computation of flux over interface based on evaluation in second region
    # @param self pointer to self
    # @param u current state
    # @return flux over boundary calculated from right side
    def compute_flux2(self, u):
        return (self.K2/self.dx)*(u[self.n+1]-u[self.n]) 
    
    ## Reference solution to functional for a givne density function
    # @param self pointer to itself
    # @return Reference solution given by extremely fine grid in time
    def solution_int(self):
        if self.u0 == 0:
            if self.K1 == 0.1 and self.K2 == 1 and self.D1 == 0.1 and self.D2 == 1:
                if self.t_end == 0.5 and self.gridsize == 50: # tol = 1.e-8
                    return 0.08084484937016052363834006655451958067715167999268
                if self.t_end == 1 and self.gridsize == 50: # tol = 1.e-8
                    return 0.15973473644067348597985755986883305013179779052734
                if self.t_end == 2 and self.gridsize == 50: # tol = 1.e-8
                    return 0.25199736873901101841255467661540023982524871826172
                print('RuntimeWarining: No reference solution available')
                return 0
            if self.K1 == 0.3 and self.K2 == 1 and self.D1 == 0.5 and self.D2 == 1:
                if self.t_end == 0.5 and self.gridsize == 50: # tol = 1.e-10
                    return 0.04583303256490661031774536127159080933779478073120
        else:
            print('RuntimeWarining: No reference solution available')
            return 0
            
            
## Class for fluid structure interaction problem for 2 coupled linear heat equations in 2 dimensions
# \n problem_type = 21
class Problem_FSI_2D(Problem_FSI):
    ## Initialization basics
    # @param self pointer to itself
    # @param gridsize gridsize for spatial discretization
    # @param t_end end time of problem
    # @param D1 heat equation related parameter
    # @param D2 heat equation related parameter
    # @param K1 heat equation related parameter
    # @param K2 heat equation related parameter
    # @param flux which subdomain to base flux calculation on
    # @param DG ignore here
    # @param u0 initial condition
    # @return None
    def __init__(self, gridsize = 50, t_end = None, D1 = 0.1, D2 = 1, K1 = 0.1, K2 = 1, flux = 1, DG = 0, u0 = 0):
        Problem_FSI.__init__(self, gridsize = gridsize, t_end = t_end, D1 = D1, D2 = D2, K1 = K1, K2 = K2, flux = flux,
                             DG = DG, u0 = u0)
                             
        ## see /misc/u0_plots.py for visualizations
        if u0 == 0: # actually has negative values, not recommended for useage
            def initial_condition(x,y):
                return (200*np.sin(np.pi*y**2)*np.sin((np.pi/2)*(x**2)))
        elif u0 == 1: # two small blobs at the far ends of each subdomain
            def initial_condition(x,y):
                return 200*np.sin(np.pi*y**2)*(x-1)**4
        elif u0 == 2: # Omega_2 and interface constant
            def initial_condition(x,y):
                if x >= 1:
                    return 200
                return 0
        elif u0 == 3: # 3x# field of smooth blobs of heat in Omega_2
            def initial_condition(x,y):
                if x >= 1:
                    return (400*(np.sin(3*np.pi*y)**2)*(np.sin(3*np.pi*x)**2))
                return 0
        elif u0 == 4: # one big blob of heat in Omega_2
            def initial_condition(x,y):
                if x >= 1:
                    return (800*(np.sin(np.pi*y)**4)*(np.sin(np.pi*x)**4))
                return 0
        elif u0 == 5: # One blob of heat in Omega_2 and another blob symmetric to the interface
            def initial_condition(x,y):
                return abs((200*np.sin(np.pi*y**2)*np.sin((np.pi/2)*(x**2))))
    
        x1 = np.linspace(0,1, self.n+2)
        xplot = np.linspace(0,2, 2*self.n+3)
        
        # only initialize the inner points
        U_0 = np.zeros((2*self.n+1, self.n))
        for i in range(2*self.n+1):
            for j in range(self.n):
                U_0[i,j] = initial_condition(xplot[i+1],x1[j+1])
        self.y0 = U_0.reshape(self.n*(2*self.n+1))
            
    ## function to compute FE Discretization matrices
    # @param self pointer to self
    # @return self.n and tons of matrices
    def compute_matrices(self):
        n, dx = self.n, self.dx
        A1g = sp.spdiags(-(self.K1/(dx**2))*np.ones(n**2), -n**2 + n, n**2, n, format='csr')
        A2g = sp.spdiags(-(self.K2/(dx**2))*np.ones(n**2), 0        , n**2, n, format='csr')
        
        Ag1 = sp.spdiags(-(self.K1/(dx**2))*np.ones(n**2), n**2 - n , n, n**2, format = 'csr')
        Ag2 = sp.spdiags(-(self.K2/(dx**2))*np.ones(n**2), 0        , n, n**2, format = 'csr')
    
        Agg1 = sp.spdiags([2*self.K1/(dx**2)*np.ones(n)] + 2*[-0.5 * self.K1/(dx**2)*np.ones(n)], [0, 1, -1], n, n, format = 'csr')
        Agg2 = sp.spdiags([2*self.K2/(dx**2)*np.ones(n)] + 2*[-0.5 * self.K2/(dx**2)*np.ones(n)], [0, 1, -1], n, n, format = 'csr')

        M1g = sp.spdiags([-self.d1/12*np.ones(n**2), self.d1/4 *np.ones(n**2)], [-n**2 + n, -n**2 + n - 1], n**2, n, format='csr')
        M2g = sp.spdiags([-self.d2/12*np.ones(n**2), self.d2/4 *np.ones(n**2)], [0, 1], n**2, n, format='csr')
        
        Mg1 = sp.spdiags([-self.d1/12*np.ones(n**2), [0]*(n**2 - n) + [self.d1/4]*n], [n**2 - n, n**2 - n - 1] , n, n**2, format='csr')
        Mg2 = sp.spdiags([-self.d2/12*np.ones(n), self.d2/4 *np.ones(n)], [0, 1], n, n**2, format='csr')
              
        Mgg1 = sp.spdiags([5*self.d1/12*np.ones(n)] + 2*[-self.d1/24*np.ones(n)], [0, 1, -1], n, n, format = 'csr')
        Mgg2 = sp.spdiags([5*self.d2/12*np.ones(n)] + 2*[-self.d2/24*np.ones(n)], [0, 1, -1], n, n, format = 'csr')

        B = sp.spdiags(4*np.ones(n),  0, n, n, format = 'csr') - \
            sp.spdiags(  np.ones((2,n)),  [-1, 1], n, n, format = 'csr') 
        A1 = sp.kron(sp.spdiags(np.ones(n), 0, n, n, format = 'csr'), B) + \
             sp.kron(sp.spdiags(np.ones((2,n)), [-1, 1], n, n, format = 'csr'), sp.spdiags(-np.ones(n), 0, n, n, format = 'csr'))
        A2 = self.K2/(dx**2)*A1
        A1 = self.K1/(dx**2)*A1
        
        N = sp.spdiags([5/6*np.ones(n)] + 2*[-1/12*np.ones(n)], [0, 1, -1], n, n, format = 'csr')
        N1 = sp.spdiags([-1/12*np.ones(n), 1/4*np.ones(n)], [0, -1], n, n, format = 'csr')
        N2 = sp.spdiags([-1/12*np.ones(n), 1/4*np.ones(n)], [0, 1], n, n, format = 'csr')
        M1 = sp.kron(sp.spdiags(np.ones(n), 0, n, n, format = 'csr'), N) + \
             sp.kron(sp.spdiags(np.ones(n), -1, n, n, format = 'csr'), N1) + \
             sp.kron(sp.spdiags(np.ones(n), 1, n, n, format = 'csr'), N2)
        M2 = self.d2*M1
        M1 = self.d1*M1
        return n, A1, A2, M1, M2, A1g, A2g, Ag1, Ag2, Agg1, Agg2, M1g, M2g, Mg1, Mg2, Mgg1, Mgg2

    ## computation of flux over interface based on evaluation in first region
    # @param self pointer to self
    # @param u current state
    # @return flux over boundary calculated from left side
    def compute_flux1(self, u):
        n = self.n
        ug = u[n*n:n*n+n]
        uaux, ugaux = np.zeros(n), np.zeros(n)
        uaux[1:n] = u[n**2 - n:n**2-1]
        ugaux[1:n] = ug[0:n-1]
        vecflux = 2*ug - uaux - ugaux
        return -self.K1*sum(vecflux)
    
    ## computation of flux over interface based on evaluation in second region
    # @param self pointer to self
    # @param u current state
    # @return flux over boundary calculated from right side
    def compute_flux2(self, u):
        n = self.n
        ug= u[n*n:n*n+n]
        uaux = np.zeros(n)
        ugaux = np.zeros(n)
        uaux[0:n-1] = u[n**2+n+1:n**2+2*n]
        ugaux[0:n-1] = ug[1:n]
        vecflux = -2*ug + uaux + ugaux
        return self.K2*sum(vecflux)
    
    ## Reference solution to functional for a givne density function
    # @param self pointer to itself
    # @return Reference solution given by extremely fine grid
    def solution_int(self):
        if self.K1 == 0.3 and self.K2 == 1 and self.D1 == 0.5 and self.D2 == 1:
            if self.t_end == 0.5 and self.gridsize == 20 and self.u0 == 0:
                return 4.17491351890692907744551121140830218791961669921875 # fixpointtol = 1e-8, tol = 1e-6
            elif self.t_end == 1 and self.gridsize == 20 and self.u0 == 1:
                return 0.49128824752213973159342685903538949787616729736328 # ???
            elif self.t_end == 1 and self.gridsize == 20 and self.u0 == 2:
                return -12.13846740900675413854514772538095712661743164062500 # tol = 1e-7, fp tol = 1e-10
            elif self.t_end == 1 and self.gridsize == 20 and self.u0 == 3:
                return -6.14816882824065924495471335831098258495330810546875 # tol = 1e-7, fp tol = 1e-10
            elif self.t_end == 1 and self.gridsize == 20 and self.u0 == 4:
                return -6.89204947019490266058028282714076340198516845703125 # tol = 1e-7, fp tol = 1e-10
            elif self.t_end == 1 and self.gridsize == 20 and self.u0 == 5:
                return 2.52286269873753887793554895324632525444030761718750 # tol = 1e-7, fp tol = 1e-10
            print('RuntimeWarining: No reference solution available')
        return 0