# -*- coding: utf-8 -*-
"""
Created on Thu Dec 10 16:09:34 2015
@author: Peter Meisrimel
credit to Patrick Farell
"""
from __future__ import division
from dolfin import *
from problem import Problem, Problem_scalar
import numpy as np

## Class for toy problem test case
# \n problem_type = 0
class Problem_Basic(Problem_scalar):
    ## initialize basics
    # @param self pointer to itself
    # @param DG Set to 1 to use DG, 0 for RK
    # @param dwr_scheme which DWR scheme to use
    # @param t_end end time of problem
    # @param func which density function to use
    # @param para_stiff stiffness parameter k in problem
    # @return None
    def __init__(self, DG = 1, dwr_scheme = 0, t_end = None, func = None, para_stiff = None):
        Problem.__init__(self, DG, dwr_scheme = dwr_scheme)
        if t_end is None: self.t_end = 2
        else:             self.t_end = t_end
        
        ## stiffness parameter
        self.k = para_stiff
        ## number indicating which density function to use
        self.func = func
        self.y0_array = np.array([1., 1.])
        self.z0_array = np.array([0., 0.])
        
        if self.DG:
            ## initial value
            self.y0_expr = Constant(self.y0_array) # initial value
            self.z0_expr = Constant(self.z0_array) # initial value
            ## Mesh being used in DG case
            self.mesh = UnitIntervalMesh(1)
            ## Function space for one component
            self.R = FunctionSpace(self.mesh, "R", 0)
            ## Function space in which solution vector operates
            self.V = self.R * self.R
            (self.v1, self.v2) = split(TestFunction(self.V))
        else:
            self.y0 = np.copy(self.y0_array)
        
    ## function to initialize problem specific run parameters and setting up weak formulations for DG case
    # @param self pointer to itself
    # @param precond which preconditioner to use for solving linear systems
    # @return None
    def initialize_vectors(self, precond = 'default'):
        Problem.initialize_vectors(self)
        if self.DG:
            (self.y1_new, self.y2_new) = split(self.y_new)
            (self.y1_old, self.y2_old) = split(self.y_old)
            (self.y1_lo, self.y2_lo) = split(self.y_lo)
            (self.l_err1, self.l_err2) = split(self.l_err)
            
            self.y1_cn = (1 - self.theta) * self.y1_old + self.theta * self.y1_new
            self.y2_cn = (1 - self.theta) * self.y2_old + self.theta * self.y2_new
            
            bcs = [DirichletBC(self.V, Constant((0.0, 0.0)), "on_boundary")]
            
            ## Crank nicolson weak formulation
            self.F = (inner(self.v1, self.y1_new - self.y1_old + self.dt * (self.y1_cn - self.y2_cn))*dx
                      + inner(self.v2, self.y2_new - self.y2_old - Constant(self.k) * self.dt * self.y2_cn)*dx)
            J = derivative(self.F, self.y_new, TrialFunction(self.V))
            self.prob = NonlinearVariationalProblem(self.F, self.y_new, bcs, J)
            self.solver = NonlinearVariationalSolver(self.prob)
            prm = self.solver.parameters
            prm['newton_solver']['linear_solver'] = 'gmres'
            prm['newton_solver']['preconditioner'] = precond
            ## Implicit Euler weak formulation for error estimation
            self.F_err = (inner(self.v1, self.y1_lo - self.y1_old + self.dt * (self.y1_lo - self.y2_lo))*dx
                          + inner(self.v2, self.y2_lo - self.y2_old - Constant(self.k) * self.dt * self.y2_lo)*dx)
            J_err = derivative(self.F_err, self.y_lo, TrialFunction(self.V))
            self.prob_err = NonlinearVariationalProblem(self.F_err, self.y_lo, bcs, J_err)
            self.solver_err = NonlinearVariationalSolver(self.prob_err)
            prm = self.solver_err.parameters
            prm['newton_solver']['linear_solver'] = 'gmres'
            prm['newton_solver']['preconditioner'] = precond
            if self.dwr_scheme: # > 0
                if self.func not in [1, 2]:
                    raise ValueError('DWR not (yet) implemented for this functional')
                else:
                    self.z_old, self.z_new = Function(self.V), Function(self.V)
                    self.z_old1, self.z_old2 = split(self.z_old)
                    self.z_new1, self.z_new2 = split(self.z_new)
                    self.z_old.interpolate(Constant((0,0)))
                    
                    self.adj_src = Function(self.V)
                    if self.func == 1:    self.adj_src.interpolate(Constant((1, 0)))
                    elif self.func == 2:    self.adj_src.interpolate(Constant((0, 1)))
                    
                    self.adj_src1, self.adj_src2 = split(self.adj_src)
                if self.dwr_scheme == 1:
                    self.F_DWR_fwd = (inner(self.v1, self.y1_new - self.y1_old + self.dt * (self.y1_new - self.y2_new))*dx
                                  + inner(self.v2, self.y2_new - self.y2_old - Constant(self.k) * self.dt * self.y2_new)*dx)
                    J_DWR_fwd = derivative(self.F_DWR_fwd, self.y_new, TrialFunction(self.V))
                    self.prob_DWR_fwd = NonlinearVariationalProblem(self.F_DWR_fwd, self.y_new, bcs, J_DWR_fwd)
                    self.solver_DWR_fwd = NonlinearVariationalSolver(self.prob_DWR_fwd)
                    prm = self.solver_DWR_fwd.parameters
                    prm['newton_solver']['linear_solver'] = 'gmres'
                    prm['newton_solver']['preconditioner'] = precond
                    
                    self.F_DWR_adj = (inner(self.z_new1 - self.z_old1 + self.dt*(-self.z_new1 + self.adj_src1), self.v1)*dx +
                                  inner(self.z_new2 - self.z_old2 + self.dt*(self.z_new1 + Constant(self.k)*self.z_new2 + self.adj_src2), self.v2)*dx)
                    J_DWR_adj = derivative(self.F_DWR_adj, self.z_new, TrialFunction(self.V))
                    self.prob_DWR_adj = NonlinearVariationalProblem(self.F_DWR_adj, self.z_new, bcs, J_DWR_adj)
                    self.solver_DWR_adj = NonlinearVariationalSolver(self.prob_DWR_adj)
                    prm = self.solver_DWR_adj.parameters
                    prm['newton_solver']['linear_solver'] = 'gmres'
                    prm['newton_solver']['preconditioner'] = precond
                elif self.dwr_scheme == 2:
                    self.solver_DWR_fwd = self.solver
                    """
                    self.F_DWR_adj = (inner(self.z_new1 - self.z_old1 + 0.5*self.dt*(-self.z_new1 - self.z_old1 + 2*self.adj_src1), self.v1)*dx +
                                      inner(self.z_new2 - self.z_old2 + 0.5*self.dt*(self.z_new1 + self.z_old1 - self.z_new2 - self.z_old2 + 2*self.adj_src2), self.v2)*dx)
                    """
                    self.F_DWR_adj = (inner(self.z_new1 - self.z_old1 + 0.5*self.dt*(-self.z_new1 - self.z_old1 + 2*self.adj_src1), self.v1)*dx +
                                      inner(self.z_new2 - self.z_old2 + 0.5*self.dt*(self.z_new1 + self.z_old1 + Constant(self.k)*(self.z_new2 + self.z_old2) + 2*self.adj_src2), self.v2)*dx)
                    J_DWR_adj = derivative(self.F_DWR_adj, self.z_new, TrialFunction(self.V))
                    self.prob_DWR_adj = NonlinearVariationalProblem(self.F_DWR_adj, self.z_new, bcs, J_DWR_adj)
                    self.solver_DWR_adj = NonlinearVariationalSolver(self.prob_DWR_adj)
                    prm = self.solver_DWR_adj.parameters
                    prm['newton_solver']['linear_solver'] = 'gmres'
                    prm['newton_solver']['preconditioner'] = precond
                else:
                    raise ValueError('No DWR scheme defined for dwr_scheme = {}'.format(self.dwr_scheme))
    
    ## density function
    # @param self pointer to itself
    # @param z input vector
    # @param val if True, converts result to float, else returns vector, default = False
    # @param force force j(z) to be treated on a non DG basis, needed for reference solutions in DG case
    # @return result of evaluating vector in density function
    def j(self, z, val = False, force = False):
        if self.DG and not force:
            if self.func == 1:    y = split(z)[0]
            elif self.func == 2:  y = split(z)[1]
            elif self.func == 10: y = split(z)[0] * split(z)[1]
            elif self.func == 21: y = split(z)[0]**2
            elif self.func == 22: y = split(z)[1]**2
            elif self.func == 31: y = split(z)[0]**3
            elif self.func == 32: y = split(z)[1]**3
            else:
                raise RuntimeError('invalid functional for this type of problem')
            if val:
                return assemble(y*dx)
            else:
                return y
        else:
            if self.func == 1:    y = z[0]
            elif self.func == 2:  y = z[1]
            elif self.func == 10: y = z[0] * z[1]
            elif self.func == 21: y = z[0]**2
            elif self.func == 22: y = z[1]**2
            elif self.func == 31: y = z[0]**3
            elif self.func == 32: y = z[1]**3
            else:
                raise RuntimeError('invalid functional for this type of problem')
            return y
    
    ## function to get pointer to right-hand side
    # @param self pointer to itself
    # @return pointer to right-hand side function
    def get_rhs(self):
        def rhs(t, y):
            return np.array([-y[0] + y[1], self.k*y[1]])
        return rhs
        
    ## Construct DWR estimatese_abs = file_read['use_abs'],
    # @param self pointer to itself
    # @param solns forward solution
    # @param adj_coarse adjoint solution on coarse grid
    # @param adj_fine adjoint solution on fine grid
    # @return estimates
    def get_dwr_est(self, solns, adj_coarse, adj_fine):
        if self.dwr_scheme == 1:
            raise ValueError('No DWR scheme defined for dwr_scheme = {}'.format(self.dwr_scheme))
        elif self.dwr_scheme == 2:
            times_coarse = adj_coarse.keys()
            times_coarse.sort()
            
            times_fine = adj_fine.keys()
            times_fine.sort()
            
            ntimesteps = len(times_coarse) - 1
            est = np.zeros(ntimesteps)
            
            u10, u20 = solns[times_coarse[0]].vector().array()
            zc10, zc20 = adj_coarse[times_coarse[0]].vector().array()
            zf10, zf20 = adj_fine[times_fine[0]].vector().array()
            for t in range(ntimesteps):
                # acess values from vectors directly
                u11, u21 = solns[times_coarse[t+1]].vector().array()
                zc11, zc21 = adj_coarse[times_coarse[t+1]].vector().array()
                zf11, zf21 = adj_fine[times_fine[2*t + 1]].vector().array()
                zf12, zf22 = adj_fine[times_fine[2*t + 2]].vector().array()
                dt = float(times_coarse[t + 1] - times_coarse[t])
                
                ut1 = (u10 - u11)/dt
                res_1_left = ut1 + u10 - u20
                res_1_right = ut1 + u11 - u21
                res_1_mid = (res_1_left + res_1_right)/2
                z1_left = zf10 - zc10
                z1_right = zf12 - zc11
                z1_mid = zf11 - (zc10 + zc11)/2
                
                part_1_left = res_1_left*z1_left
                part_1_right = res_1_right*z1_right
                part_1_mid = res_1_mid*z1_mid
                
                ut2 = (u20 - u21)/dt
                res_2_left = ut2 - self.k*u20
                res_2_right = ut2 - self.k*u21
                res_2_mid = (res_2_left + res_2_right)/2
                z2_left = zf20 - zc20
                z2_right = zf22 - zc21
                z2_mid = zf21 - (zc20 + zc21)/2
                
                part_2_left = res_2_left*z2_left
                part_2_right = res_2_right*z2_right
                part_2_mid = res_2_mid*z2_mid
                
                est[t] = (abs(dt/4*(part_1_left + 2*part_1_mid + part_1_right) + 
                          dt/4*(part_2_left + 2*part_2_mid + part_2_right)))
                # shift of values by one
                u10, u20, zc10, zc20, zf10, zf20 = u11, u21, zc11, zc21, zf12, zf22
            return est
        else:
            raise ValueError('No DWR scheme defined for dwr_scheme = {}'.format(self.dwr_scheme))
    
    ## Evaluate solution for a given point in time, does not include density function
    # @param self pointer to itself
    # @param t point in time to evaluate solution, default = self.t_end
    # @param y_loc starting y for local problem, only relevant for local == True
    # @return Value of exact solution at given point
    def solution_point(self, t = None, y_loc = None):
        from numpy import exp
        if t is None    : t = self.t_end
        if y_loc is None: u1, u2 = self.y0_array
        else:             u1, u2 = y_loc
        
        if self.k != -1: return np.array([exp(-t)*((1 + self.k)*u1 + (exp((self.k + 1)*t) - 1)*u2)/(self.k + 1), exp(t*self.k)*u2])
        else:            return np.array([exp(-t)*(u1 + t*u2), exp(-t)*u2])
            
    ## Evaluate solution of functional between any two given point for density function self.func, see mathematic file
    # @param self pointer to itself
    # @param t_0 lower boundary, default = 0
    # @param t_end upper boundary, default = self.t_end
    # @return Value of funtional
    def solution_int(self, t_0 = None, t_end = None):
        from numpy import exp
        if t_0 is None:   t_0 = 0
        if t_end is None: t_end = self.t_end
        u1, u2 = self.y0_array
        k = self.k
        if self.func == 1:
            def res(t):
                if self.k != -1: return (exp(-t)*(-k**2*u1 + exp((1 + k)*t)*u2 + k*(-u1 + u2)))/(k*(1 + k))
                else:            return -exp(-t)*(u1 + u2 + t*u2)
        elif self.func == 21:
            def res(t):
                if self.k != -1: return (exp(-2*t)*(-k**4*u1**2 - k**3*u1*(u1 - 2*u2) - exp(2*(1 + k)*t)*u2**2 + k**2*(u1**2 + 
                                         4*exp((1 + k)*t)*u1*u2 - u2**2) + k*(u1**2 + 2*(-1 + 2*exp((1 + k)*t))*u1*u2 + (1 - 4*exp((1 + 
                                         k)*t) + exp(2*(1 + k)*t))*u2**2)))/(2*(-1 + k)*k*(1 + k)**2)
                else:            return -(1/4)*exp(-2*t)*(2*u1**2 + (1 + 2*t + 2*t**2)*u2**2 + 2*u1*(u2 + 2*t*u2))
        elif self.func == 31:
            def res(t):
                if self.k != -1: return (1/(3*(-2 + k)*k*(1 + k)**3*(-1 + 2*k)))*exp(-3*t)*(-2*k**6*u1**3 - k**5*u1**2*(u1 - 6*u2) + 
                                         2*exp(3*(1 + k)*t)*u2**3 + k**4*u1*(7*u1**2 + 3*(-1 + 6*exp((1 + k)*t))*u1*u2 - 6*u2**2) + 
                                         k**3*(7*u1**3 + 9*(-2 + 3*exp((1 + k)*t))*u1**2*u2 + 9*(1 - 4*exp((1 + k)*t) + exp(2*(1 + 
                                         k)*t))*u1*u2**2 + 2*u2**3) - k**2*(u1**3 + 3*u1**2*u2 + 9*(-1 + 2*exp((1 + k)*t) + exp(2*(1 + 
                                         k)*t))*u1*u2**2 - (-5 + 18*exp((1 + k)*t) - 9*exp(2*(1 + k)*t) + 2*exp(3*(1 + k)*t))*u2**3) - 
                                         k*(2*u1**3 + 3*(-2 + 3*exp((1 + k)*t))*u1**2*u2 + 6*(1 - 3*exp((1 + k)*t) + 3*exp(2*(1 + 
                                         k)*t))*u1*u2**2 + (-2 + 9*exp((1 + k)*t) - 18*exp(2*(1 + k)*t) + 5*exp(3*(1 + k)*t))*u2**3))
                else:            return 1/27*exp(-3*t)*(-9*u1**3 - 3*(2 + 6*t + 9*t**2)*u1*u2**2 - (2 + 6*t + 9*t**2 + 9*t**3)*u2**3 - 
                                        9*u1**2*(u2 + 3*t*u2))
        elif self.func == 2:
            def res(t):          return (exp(k*t)*u2)/k
        elif self.func == 22:
            def res(t):          return (exp(2*k*t)*u2**2)/(2*k)
        elif self.func == 32:
            def res(t):          return (exp(3*k*t)*u2**3)/(3*k)
        elif self.func == 10:
            def res(t):
                if self.k != -1: return (exp((-1 + k)*t)*u2*(2*k**2*u1 - exp((1 + k)*t)*u2 + k*(2*u1 + (-2 + exp((1 + 
                                         k)*t))*u2)))/(2*k*(-1 + k**2))
                else:            return -(1/4)*exp(-2*t)*u2*(2*u1 + u2 + 2*t*u2)
        else:                    raise ValueError('No reference solution available')
        return res(t_end) - res(t_0)