# -*- coding: utf-8 -*-
"""
Created on Fri Oct 20 14:45:27 2017
@author: Peter Meisrimel
"""
from __future__ import division
import numpy as np
from dolfin import *
from problem import Problem, Problem_2D

## Class for linear heat equation problem with source term
# \n problem_type = 12
class Problem_heat_source_adv(Problem_2D):
    time_dep = True
    ## Initialization basics
    # @param self pointer to itself
    # @param dwr_scheme which DWR scheme to use
    # @param func which density function to use
    # @param gridsize for spatial discretization
    # @param domain which domain to use for density function
    # @param err_est_type type of error estimation to use, mostly spatial based things
    # @param t_end end time of problem
    # @param direction direction of advection, 1 = right, 0 = left
    # @return None
    def __init__(self, dwr_scheme = 0, func = None, gridsize = 32, domain = 1, err_est_type = 1,
                 t_end = 6, direction = 1):
        Problem.__init__(self, DG = 1, dwr_scheme = dwr_scheme)
        
        ## end time
        self.t_end = t_end
        
        ## initial value
        self.y0_expr = Expression("1") # initial values
        self.z0_expr = Expression("0")
        
        # heat source specific parameterc
        ## maximum value of heat prodocution
        self.src_m = 5.
        ## wind up&down time
        self.src_s = 1.
        ## heat production stop time
        self.src_t_s = 2.
        ## heat source term for weak formulation for Crank Nicolson, IE uses only f1
        self.f1 = Expression('m*pow(t/s,3)', t = 0, m=self.src_m, s=self.src_s, degree = 2)
        self.f2 = Expression('m*pow(t/s,3)', t = 0, m=self.src_m, s=self.src_s, degree = 2)
        
        ## Size of grid, number of cells = (2*gridsize)*gridsize*2
        self.gridsize = gridsize
        ## Mesh 
        self.mesh = RectangleMesh(Point(0,0), Point(3,1), 3*self.gridsize, self.gridsize)
        ## Function space for solution vector
        self.V = FunctionSpace(self.mesh, "CG", 1)
        ## Function space for velocity
        self.V_velo = VectorFunctionSpace(self.mesh, "CG", 2)
        ## subdomain scheme used for functional
        self.domain = domain
        
        ## direction of velocity
        self.direction = direction
        
        self.n_facet = FacetNormal(self.mesh)

        # create measures for relevant subdomains
        colours = CellFunction("size_t", self.mesh) # Colouring for mesh to identify various subdomains
        colours.set_all(0) # default colour
        
        # heat source
        heat_src = CompiledSubDomain("x[0] >= 0.25 && x[0] <= 0.75 && x[1] >= 0.25 && x[1] <= 0.75") # identify subdomain by condition
        heat_src.mark(colours, 2) # change colour of domain
        self.dx_src = Measure("dx", subdomain_data=colours)(2) # define domain in the needed way for function definition

        if self.domain != 0: # full domain
            if self.domain == 1: # small square
                subdom = CompiledSubDomain("x[0] >= 2.25 && x[0] <= 2.75 && x[1] >= 0.25 && x[1] <= 0.75")
            subdom.mark(colours, 1) # change colour of domain
            ## used for integration of subdomain
            self.dx_obs = Measure("dx", subdomain_data = colours)(1) # define domain in the needed way for function definition
        else:
            self.dx_obs = dx
        self.dx_cont = self.dx_obs
            
        ## Identifier for density function to be used
        self.func = func
        
    ## Density function
    # @param self pointer to itself
    # @param z Vector to evaluate
    # @param val Set to True get return as float
    # @param force Ignore here
    # @return Result of evaluating z with density function
    def j(self, z, val = False, force = False):
        if self.func == 1:
            y = z/self.t_end
        else:
            raise ValueError('invalid functional for this type of problem')
        if val:
            return assemble(y*self.dx_obs)
        else:
            return y
        
    ## Functon to initialize running parameters and setting up variational formulation
    # @param self pointer to itself
    # @param precond which preconditioner to use for solving linear systems
    # @return None
    def initialize_vectors(self, precond = 'default'):
        Problem.initialize_vectors(self)
        
        ## y used in Crank Nicolson
        self.y_cn = self.theta * self.y_new + (1 - self.theta) * self.y_old
        ## Test function for weak formulation
        self.v = TestFunction(self.V)
        if self.direction == 1:
            self.velocity = interpolate(Expression(('1', '0')), self.V_velo)
        elif self.direction == -1:
            self.velocity = interpolate(Expression(('-1', '0')), self.V_velo)
        
        self.heat_cond = Constant(0.01)
        self.outflow = Constant(0.15)
        self.velo = Constant(0.5)
        ## Variational formulation for Crank Nicolson
        self.F = (inner(self.v, self.y_new - self.y_old)*dx
                  + self.dt*self.heat_cond*inner(grad(self.y_cn), grad(self.v))*dx
                  - 0.5*self.dt*inner(self.f1 + self.f2, self.v)*self.dx_src
                  + self.velo*self.dt*dot(grad(self.y_cn), self.velocity)*self.v*dx
                  + self.dt*self.heat_cond*self.outflow*inner(self.y_cn, self.v)*ds)
        J = derivative(self.F, self.y_new, TrialFunction(self.V))
        self.prob = NonlinearVariationalProblem(F = self.F, u = self.y_new, J = J)
        self.solver = NonlinearVariationalSolver(self.prob)
        prm = self.solver.parameters
        prm['newton_solver']['linear_solver'] = 'gmres'
        prm['newton_solver']['preconditioner'] = precond
        ## Boundaries for Crank Nicolson
        
        ## Variational formulation for Implicit Euler
        self.F_err = (inner(self.v, self.y_lo - self.y_old)*dx
                      + self.dt*self.heat_cond*inner(grad(self.y_lo), grad(self.v))*dx
                      - self.dt*inner(self.f2, self.v)*self.dx_src
                      + self.velo*self.dt*dot(grad(self.y_lo), self.velocity)*self.v*dx
                      + self.dt*self.heat_cond*self.outflow*inner(self.y_lo, self.v)*ds)
        J_err = derivative(self.F_err, self.y_lo, TrialFunction(self.V))
        self.prob_err = NonlinearVariationalProblem(F = self.F_err, u = self.y_lo, J = J_err)
        self.solver_err = NonlinearVariationalSolver(self.prob_err)
        prm = self.solver_err.parameters
        prm['newton_solver']['linear_solver'] = 'gmres'
        prm['newton_solver']['preconditioner'] = precond
        
        if self.dwr_scheme: # > 0
            if self.func not in [1]:
                raise ValueError('DWR not (yet) implemented for this functional')
            else:
                self.z_old, self.z_new = Function(self.V), Function(self.V)
                self.z_old.interpolate(Constant(0))
                
                self.adj_src = Function(self.V)
                if self.func == 1:    self.adj_src.interpolate(Constant(1/self.t_end))
                
                self.src_weight = interpolate(src_weight(), self.V)
                
                class LeftBoundary(SubDomain):
                    def inside(self, x, on_boundary):
                        return on_boundary and near(x[0], 0)
                            
                class LowerBoundary(SubDomain):
                    def inside(self, x, on_boundary):
                        return on_boundary and near(x[1], 0)
                    
                class RightBoundary(SubDomain):
                    def inside(self, x, on_boundary):
                        return on_boundary and near(x[0], 3)
                            
                class TopBoundary(SubDomain):
                    def inside(self, x, on_boundary):
                        return on_boundary and near(x[1], 1)
                        
                left = LeftBoundary()
                lower = LowerBoundary()
                right = RightBoundary()
                top = TopBoundary()
                        
                boundaries = FacetFunction("size_t", self.mesh)
                boundaries.set_all(0)
                left.mark(boundaries, 1)
                lower.mark(boundaries, 2)
                right.mark(boundaries, 3)
                top.mark(boundaries, 4)
                
                ds_self = Measure('ds', domain = self.mesh, subdomain_data=boundaries)
            if self.dwr_scheme == 1:
                self.F_DWR_fwd = (inner(self.v, self.y_new - self.y_old)*dx -
                                  self.dt*self.heat_cond*inner(self.v, self.f2)*self.dx_src + 
                                  self.velo*self.dt*dot(grad(self.y_new), self.velocity)*self.v*dx + 
                                  self.dt*self.heat_cond*self.outflow*inner(grad(self.v), grad(self.y_new))*dx)
                J_DWR_fwd = derivative(self.F_DWR_fwd, self.y_new, TrialFunction(self.V))
                self.prob_DWR_fwd = NonlinearVariationalProblem(F = self.F_DWR_fwd, u = self.y_new, J = J_DWR_fwd)
                self.solver_DWR_fwd = NonlinearVariationalSolver(self.prob_DWR_fwd)
                prm = self.solver_DWR_fwd.parameters
                prm['newton_solver']['linear_solver'] = 'gmres'
                prm['newton_solver']['preconditioner'] = precond
                
                self.F_DWR_adj = (inner(self.z_new - self.z_old, self.v)*dx -
                                  self.dt*self.heat_cond*inner(grad(self.z_new), grad(self.v))*dx +
                                  self.dt*inner(self.adj_src, self.v)*self.dx_obs +
                                  self.dt*self.velo*dot(self.velocity, grad(self.z_new))*self.v*dx -
                                  self.dt*self.velo*dot(self.velocity, grad(self.z_new))*self.v*ds_self(1) -
                                  self.dt*self.velo*dot(self.velocity, grad(self.z_new))*self.v*ds_self(3))
                                  
                J_DWR_adj = derivative(self.F_DWR_adj, self.z_new, TrialFunction(self.V))
                self.prob_DWR_adj = NonlinearVariationalProblem(F = self.F_DWR_adj, u = self.z_new, J = J_DWR_adj)
                self.solver_DWR_adj = NonlinearVariationalSolver(self.prob_DWR_adj)
                prm = self.solver_DWR_adj.parameters
                prm['newton_solver']['linear_solver'] = 'gmres'
                prm['newton_solver']['preconditioner'] = precond
            elif self.dwr_scheme == 2:
                self.solver_DWR_fwd = self.solver

                self.F_DWR_adj = (inner(self.z_new - self.z_old, self.v)*dx -
                                  0.5*self.dt*self.heat_cond*inner(grad(self.z_new + self.z_old), grad(self.v))*dx +
                                  self.dt*inner(self.adj_src, self.v)*self.dx_obs +
                                  self.direction*0.5*self.dt*self.velo*dot(self.velocity, grad(self.z_new + self.z_old))*self.v*dx -
                                  0.5*self.velo*self.dt*inner(self.z_new + self.z_old, self.v)*ds_self(1) -
                                  0.5*self.velo*self.dt*inner(self.z_new + self.z_old, self.v)*ds_self(3) - 
                                  0.5*self.outflow*self.heat_cond*self.dt*inner(self.z_new - self.z_old, self.v)*ds)
                J_DWR_adj = derivative(self.F_DWR_adj, self.z_new, TrialFunction(self.V))
                self.prob_DWR_adj = NonlinearVariationalProblem(F = self.F_DWR_adj, u = self.z_new, J = J_DWR_adj)
                self.solver_DWR_adj = NonlinearVariationalSolver(self.prob_DWR_adj)
                prm = self.solver_DWR_adj.parameters
                prm['newton_solver']['linear_solver'] = 'gmres'
                prm['newton_solver']['preconditioner'] = precond
            else:
                raise ValueError('No DWR scheme defined for dwr_scheme = {}'.format(self.dwr_scheme))
                
    ## DWR error estimation
    # @param self pointer to itself
    # @param solns forward solution
    # @param adj_coarse adjoint solution of coarse grid
    # @param adj_fine adjoint solution of fine grid
    # @return Error estimate
    def get_dwr_est(self, solns, adj_coarse, adj_fine):
        if self.dwr_scheme == 1:
            raise ValueError('No DWR scheme defined for dwr_scheme = {}'.format(self.dwr_scheme))
        elif self.dwr_scheme == 2: # CG1
            times_coarse = adj_coarse.keys()
            times_coarse.sort()
            
            times_fine = adj_fine.keys()
            times_fine.sort()
            
            ntimesteps = len(times_coarse) - 1
            est = np.zeros(ntimesteps)
            
            u0 = solns[times_coarse[0]]
            
            div0 = self.heat_cond*div(grad(u0))
            adv0 = self.direction*self.velo*dot(grad(u0), self.velocity)
            zc0 = adj_coarse[times_coarse[0]];
            f0 = Function(self.V); self.f1.t = times_coarse[0]; f0.interpolate(self.f1)
            f0 = f0*self.src_weight
            zf0 = adj_fine[times_fine[0]];
            for t in range(ntimesteps):
                u1 = solns[times_coarse[t + 1]]
                div1 = self.heat_cond*div(grad(u1))
                adv1 = self.direction*self.velo*dot(grad(u1), self.velocity)
                f1 = Function(self.V); self.f1.t = times_coarse[t + 1]; f1.interpolate(self.f1)
                f1 = f1*self.src_weight
                zc1 = adj_coarse[times_coarse[t + 1]]
                zf1 = adj_fine[times_fine[2*t + 1]]; zf2 = adj_fine[times_fine[2*t + 2]];
                dt = Constant(times_coarse[t + 1] - times_coarse[t])

                ut = (u1 - u0)/dt
                res_left = ut - div0 + adv0 - f0
                res_right = ut - div1 + adv1 - f1
                res_mid = (res_left + res_right)/2
                zz_left = zf0 - zc0
                zz_right = zf2 - zc1
                zz_mid = zf1 - (zc0 + zc1)/2

                part_left = assemble(res_left*zz_left*dx)
                part_right = assemble(res_right*zz_right*dx)
                part_mid = assemble(res_mid*zz_mid*dx)
                
                est[t] = abs(dt/4*(part_left + 2*part_mid + part_right))
                
                u0, div0, zc0, f0, zf0 = u1, div1, zc1, f1, zf2
            return est
        else:
            raise ValueError('No DWR scheme defined for dwr_scheme = {}'.format(self.dwr_scheme))
    
    ## Reference solution to functional for a given density function
    # @param self pointer to itself
    # @return Reference solution given by extremely fine grid
    def solution_int(self):
        res = 0
        if self.domain == 1:
            if self.t_end == 6. and self.gridsize == 32 and self.direction == 1: # tol = 1.e-6
                if self.func == 1: res = 1.02243287066784671601737954915734007954597473144531 # tol = 1e-6, new dt0
            if self.t_end == 3. and self.gridsize == 32 and self.direction == -1: # tol = 1.e-6
                if self.func == 1: res = 0.23715659102035066552183195653924485668540000915527 # tol = 1e-9, new dt0
        if res == 0:
            print('RuntimeWarning: No reference solution available')
        return res
        
class src_weight(Expression):
    ## initialization function, desired region is a square and marked by boundary values
    # @param self pointer to self
    # @param left_lim value for left boundary
    # @param right_lim value for right boundary
    # @param lower_lim value for lower boundary
    # @param upper_lim value for upper boundary
    def __init__(self, left_lim = 0.25, right_lim = 0.75, lower_lim = 0.25, upper_lim = 0.75,):
        self.left_lim = left_lim
        self.right_lim = right_lim
        
        self.lower_lim = lower_lim
        self.upper_lim = upper_lim
    
    ## function that is called when evaluating a single point/cell of the grid
    # @param self pointer to self
    # @param value output value, has to be assigned via value[0]
    # @param x coordinates of point, accessed via x[0], x[1]
    # @param ufc_cell required
    # @return None, assign result to value[0]
    def eval_cell(self, value, x, ufc_cell):
        if self.left_lim <= x[0] <= self.right_lim and self.lower_lim <= x[1] <= self.upper_lim:
            value[0] = 1
        else:
            value[0] = 0