# -*- coding: utf-8 -*-
"""
Created on Thu Jun 15 16:38:46 2017
@author: Peter Meisrimel
"""
from __future__ import division
from dolfin import *
from problem import Problem, Problem_scalar
import numpy as np

## Verification est problem \f$ u'(t) = -u(t), \quad u(0) = 1 \f$.
# \n problem type = -1
class Problem_verification(Problem_scalar):
    ## init function
    # @param self pointer to self
    # @param DG Set to 1 for DG, 0 for RK case
    # @param dwr_scheme which DWR scheme to use, 0 for non DWR case
    # @param t_end end time of problem
    # @param func which density function to use
    # @return None
    def __init__(self, DG = 1, dwr_scheme = 0, t_end = None, func = None):
        Problem.__init__(self, DG, dwr_scheme = dwr_scheme)
        
        if t_end is None: self.t_end = 1
        else:             self.t_end = t_end
        self.func = func
        self.y0_array = np.array([1.])
        
        if self.DG:
            self.y0_expr = Constant(1) # initial value
            self.mesh = UnitIntervalMesh(1)
            if self.DG == 2:
                self.V = FunctionSpace(self.mesh, 'DG', 0)
            else:
                self.V = FunctionSpace(self.mesh, "R", 0)
        else:
            self.y0 = np.copy(self.y0_array)
            
    ## Function to initialize vectors and set up variational formulation
    # @param self pointer to self
    # @param precond which preconditioner to use for solving linear systems
    # @return None
    def initialize_vectors(self, precond = 'default'):
        Problem.initialize_vectors(self)
        if self.DG:
            self.y_cn = (1 - self.theta) * self.y_old + self.theta * self.y_new
            self.v = TestFunction(self.V)
            
            bcs = [DirichletBC(self.V, 0.0, "on_boundary")]
            # CN, higher order
            self.F = inner(self.v, self.y_new - self.y_old + self.dt * self.y_cn)*dx
            J = derivative(self.F, self.y_new, TrialFunction(self.V))
            self.prob = NonlinearVariationalProblem(self.F, self.y_new, bcs, J)
            self.solver = NonlinearVariationalSolver(self.prob)
            prm = self.solver.parameters
            prm['newton_solver']['linear_solver'] = 'gmres'
            prm['newton_solver']['preconditioner'] = precond
            # IE, low order
            self.F_err = inner(self.v, self.y_lo - self.y_old + self.dt * self.y_lo)*dx
            J_err = derivative(self.F_err, self.y_lo, TrialFunction(self.V))
            self.prob_err = NonlinearVariationalProblem(self.F_err, self.y_lo, bcs, J_err)
            self.solver_err = NonlinearVariationalSolver(self.prob_err)
            prm = self.solver_err.parameters
            prm['newton_solver']['linear_solver'] = 'gmres'
            prm['newton_solver']['preconditioner'] = precond
            if self.dwr_scheme: # > 0
                if self.dwr_scheme == 1:
                    self.F_DWR = inner(self.v, self.y_new - self.y_old + self.dt * self.y_new)*dx
                    J_dwr = derivative(self.F_err, self.y_new, TrialFunction(self.V))
                    self.prob_dwr = NonlinearVariationalProblem(self.F_DWR, self.y_new, bcs, J_dwr)
                    self.solver_DWR = NonlinearVariationalSolver(self.prob_DWR)
                    prm = self.solver_DWR.parameters
                    prm['newton_solver']['linear_solver'] = 'gmres'
                    prm['newton_solver']['preconditioner'] = precond
                elif self.dwr_scheme == 2:
                    self.solver_DWR = self.solver
                else:
                    raise ValueError('No DWR scheme defined for dwr_scheme = {}'.format(self.dwr_scheme))
                    
    ## density function :\f$ j(u) \f$
    # @param self pointer to self
    # @param z input function value
    # @param val if True, converts results to discrete value
    # @param force force j(z) to be treated on a non DG basis, needed for reference solutions in DG case
    # @return Value for j(z)
    def j(self, z, val = False, force = False):
        if self.DG:
            if   self.func == 1:  y = z
            elif self.func == 21: y = z**2
            elif self.func == 31: y = z**3
            else:
                raise RuntimeError('invalid functional for this type of problem')
            if val: return assemble(y*dx)
        else:
            if   self.func == 1:  y = z[0]
            elif self.func == 21: y = z[0]**2
            elif self.func == 31: y = z[0]**3
            else:
                raise RuntimeError('invalid functional for this type of problem')
        return y
     
    ## Function to get a pointer to rhs function
    # @param self pointer to self
    # @return rhs function
    def get_rhs(self):
        def rhs(t, y):
            return np.array([-y[0]])
        return rhs
    
    ## Function to estimate the DWR error
    # @param self pointer to self
    # @param mesh mesh used for time discretization
    # @param solns Forward in time solution
    # @param adjsolns Backward in time solution
    # @return Estimate for DWR error
    def get_dwr_est(self, mesh, solns, adjsolns):
        if self.dwr_scheme == 1: # DG0
            est = []
            for cell in cells(mesh):
                dt = cell.diameter()
                (t, t_new) = cell.get_vertex_coordinates()
                t, t_new = min(t, t_new), max(t, t_new)
                est.append(abs(0.5*dt*solns[t_new].vector()[0]*(adjsolns[t].vector()[0] - adjsolns[t_new].vector()[0])))
            return est
        else:
            raise ValueError('DWR estimator not implemented')
    
    ## Returns solution of a single point in time
    # @param self pointer to self
    # @param t point in time to evalute solution
    # @param local set to True for solution to local problem, requires y_local != None
    # @param dt timestep done, only needed in local case
    # @param y_loc starting y for local problem, only relevant for local == True
    # @return Value of exact solution at given point
    def solution_point(self, t = None, local = False, dt = None, y_loc = None):
        from numpy import exp
        if not local:
            if t is None: t = self.t_end
            return np.array([exp(-t)])
        else: # local problem
            return exp(dt)*y_loc
    
    ## Return solution of integral over time\n
    # @param self pointer to self
    # @param t_0 lower boundary point, default = 0
    # @param t_end upper boundary point, default = self.t_end
    # @return Value of solution
    def solution_int(self, t_0 = None, t_end = None):
        from numpy import exp
        if t_0 is None:   t_0 = 0
        if t_end is None: t_end = self.t_end
        if self.func == 1:
            def res(t):  return -exp(-t)
        elif self.func == 21:
            def res(t):  return -exp(-2*t)*0.5
        elif self.func == 31:
            def res(t):  return -exp(-3*t)/3.
        return res(t_end) - res(t_0)

## Verification est problem \f$ u'(t) = -u(t) + t^2, \quad u(0) = 1 \f$
# \n problem_type = -2
class Problem_verification_time(Problem_verification):
    time_dep = True
    ## Parameter initialisation function
    # @param self pointer to self
    # @param DG Set to 1 in DG case, 0 for RK
    # @param dwr_scheme which DWR method to use, 0 for non DWR case
    # @param precond which preconditioner to use for solving linear systems
    # @return None
    def initialize_vectors(self, DG = 1, dwr_scheme = 0, precond = 'default'):
        Problem.initialize_vectors(self)
        if self.DG:
            self.y_cn = (1 - self.theta) * self.y_old + self.theta * self.y_new
            self.v = TestFunction(self.V)
            
            # rhs
            self.f1 = Expression('pow(t, 2)', t = 0, degree = 1)
            self.f2 = Expression('pow(t, 2)', t = 0, degree = 1)
            
            bcs = [DirichletBC(self.V, 0.0, "on_boundary")]
            # CN, higher order
            self.F = (inner(self.v, self.y_new - self.y_old + self.dt * self.y_cn)*dx
                     - 0.5*self.dt*inner(self.f1 + self.f2, self.v)*dx)
            J = derivative(self.F, self.y_new, TrialFunction(self.V))
            self.prob = NonlinearVariationalProblem(self.F, self.y_new, bcs, J)
            self.solver = NonlinearVariationalSolver(self.prob)
            prm = self.solver.parameters
            prm['newton_solver']['linear_solver'] = 'gmres'
            prm['newton_solver']['preconditioner'] = precond
            # IE, low order
            self.F_err = (inner(self.v, self.y_lo - self.y_old + self.dt * self.y_lo)*dx
                          - self.dt*inner(self.f2, self.v)*dx)
            J_err = derivative(self.F_err, self.y_lo, TrialFunction(self.V))
            self.prob_err = NonlinearVariationalProblem(self.F_err, self.y_lo, bcs, J_err)
            self.solver_err = NonlinearVariationalSolver(self.prob_err)
            prm = self.solver_err.parameters
            prm['newton_solver']['linear_solver'] = 'gmres'
            prm['newton_solver']['preconditioner'] = precond
            
            if self.dwr_scheme == 2: # adjoint run mode case
                self.solver_DWR = self.solver
                
    ## Gets right-hand side function as a pointer
    # @param self pointer to self
    # @return pointer to righ-hand side function
    def get_rhs(self):
        def rhs(t, y):
            return np.array([-y[0] + t**2])
        return rhs
    
    """ OUTDATED // UNTESTED
    ## Get DWR error estimate
    # @param self pointer to self
    # @param times list of points in time
    # @param solns forward solution
    # @param adjsolns adjoint solution
    # @return DWR estimates
    def get_dwr_est(self, times, solns, adjsolns):
        if self.dwr_scheme:
            ntimesteps = len(times) - 1
            est = np.zeros(ntimesteps)
            for timestep in range(ntimesteps):
                k = times[timestep+1] - times[timestep]
                res = np.array([solns[times[timestep+1]].vector().array() - times[timestep+1]**2])
                
                dz = adjsolns[times[timestep+1]].vector().array() - adjsolns[times[timestep]].vector().array()
                err = np.dot(res, 0.5*k*dz)
                est[timestep] = abs(err)
            return est
        else:
            raise ValueError('DWR estimator not implemented')
    """
    
    ## Returns soltuion at a given point in time
    # @param self pointer to self
    # @param t point in time, default = self.t_end
    # @return Value at given point in time
    def solution_point(self, t = None):
        if t is None:   t = self.t_end
        return np.array([-np.exp(-t) + t**2 - 2*t + 2])
    
    ## Returns soltuion of functional between any two given points, uses desity function self.func
    # @param self pointer to self
    # @param t_0 lower boundary point, default = 0
    # @param t_end upper boundary point, default = self.t_end
    # @return Value of integral for given points
    def solution_int(self, t_0 = None, t_end = None):
        if t_0 is None:   t_0 = 0
        if t_end is None: t_end = self.t_end
        if self.func == 1:
            def res(t):  return t**3/3. - t**2 + 2*t + np.exp(-t)
        elif self.func == 21:
            def res(t):  return np.exp(-2*t)*(-15 + 60*np.exp(t)*(2 + t**2) + 
                                2*np.exp(2*t)*t*(60 - 60*t + 40*t**2 - 15*t**3 + 3*t**4))/30.
        elif self.func == 31:
            def res(t):  return np.exp(-3*t)*(140 - 315*np.exp(t)*(3 - 2*t + 2*t**2) + 1260*np.exp(2*t)*(12 + 8*t + 8*t**2 + t**4) +
                                12*np.exp(3*t)*t*(280 - 420*t + 420*t**2 - 280*t**3 + 126*t**4 - 35*t**5 + 5*t**6))/420.
        return res(t_end) - res(t_0)