# -*- coding: utf-8 -*-
"""
Created on Mon Jan  9 13:22:54 2017

@author: Peter Meisrimel, Lund University
"""

from __future__ import division
import numpy as np
import pylab as pl
import itertools

## Generic parent class for various run modes
class run_simulation():
    ## Basic initialization
    # @param self pointer to self
    # @param file_read parameters read from file
    # @param legends collected legend entries if run mode requires several files
    # @return None
    def __init__(self, file_read, legends):
        self.legends = legends
        self.save_file_add = 'prob_' + str(file_read['problem_type']) + '_func_' + str(file_read['func_type'])
        if int(file_read['problem_type']) == 0:
            self.save_file_add = self.save_file_add + '_k_' + str(file_read['problem_stiff_para'])
        elif int(file_read['problem_type']) in [10, 11, 12]:
            self.save_file_add = self.save_file_add + '_dx_' + str(file_read['problem_meshsize']) + '_dom_' + str(file_read['problem_domain'])
        self.save_format = file_read['file_ending']
        self.reset_markers()
        pl.rcParams['lines.linewidth'] = 3
        pl.rcParams['lines.markersize'] = 17
    ## Basic file information to write into save file
    # @param self pointer to self
    # @param file_read parameters read from file
    # @return None
    def save_file_header(self, file_read):
        self.myfile.write('PROBLEM_TYPE   : {}'.format(file_read['problem_type']))
        self.myfile.write('\nSTIFFNESS_PARA : {}'.format(file_read['problem_stiff_para']))
        self.myfile.write('\nGRIDSIZE       : {}'.format(file_read['problem_meshsize']))
        self.myfile.write('\nT_END          : {}'.format(file_read['problem_t_end']))
        self.myfile.write('\nDOMAIN         : {}'.format(file_read['problem_domain']))
        self.myfile.write('\nFUNC_TYPE      : {}'.format(file_read['func_type']))
        if file_read['problem_type'] in (20, 21):
            self.myfile.write('\nD1             : {}'.format(file_read['fsi_d1']))
            self.myfile.write('\nD2             : {}'.format(file_read['fsi_d2']))
            self.myfile.write('\nK1             : {}'.format(file_read['fsi_k1']))
            self.myfile.write('\nK2             : {}'.format(file_read['fsi_k2']))
            self.myfile.write('\nFIXPOINT TOL   : {}'.format(file_read['fsi_tol_fixpoint']))
            
    ## Data to write into file, is done by passing a tuples
    # @param self pointer to self
    # @param args Tuples of data to write into file, format : (String, (List|Array))*
    # @return None
    def save_file_write(self, *args):
        for i in args:
            if type(i[1]) in (list, np.ndarray):
                self.myfile.writelines('\n' + i[0] + ' = ' + str(list(i[1])))
            else:
                self.myfile.writelines('\n' + i[0] + ' = %10.50f' % i[1])
    
    ## Function to end writing to file process and close the file
    # @param self pointer to self
    # @return None
    def save_file_end(self):
        self.myfile.close()
        
    ## Function to (re)set markers and colors for plots to cycle
    # @param self pointer to self
    # @return None
    def reset_markers(self):
        self.markers = itertools.cycle('o^D*s')
        self.colors = itertools.cycle('gbrckmy')
        self.linestyles = itertools.cycle(['-', '--'])
    
    ## Function to set the ylimits according to the data and limit by 1e-16
    # @param self pointer to self
    # @param args axes objects to perform limiting for
    # @return None
    def ylim(self, *args):
        for ax in args:
            up, low = -np.Inf, np.Inf
            for vals in ax.lines:
                up = max(up, max(vals.get_ydata()))
                low = min(low, min(vals.get_ydata()))
            ax.set_ylim(bottom = 10**(max(np.floor(np.log10(low)), -16)), top = 10**(np.ceil(np.log10(up))))
        
    ## main function for executing a run
    # @param self pointer to self
    # @param file_read parameters read from file
    # @return None
    def simulate(self, file_read):
        pass
    
    ## that one ugly functions in which all the values and functions are exchanged violating common access conventions of class paramters and functions
    # @param self pointer to self
    # @param file_read parameters read from file
    # @param tol tolerance used to timestep controller
    # @param DG choose if to use DG scheme or not, default will take parameter read from file
    # @param dwr_scheme scheme to be used in DWR, 0 in non DWR case
    # @param diagnostics indicator if Time integration is to create extra statistics during run
    # @param dt_cont_type used to overwrite dt_cont_type given by input file
    # @param time_int_class which Time integration type to use, only needed to be set 1 if using run_global_error
    # @param err_est_type type or error estimator used
    # @return Time integration class, initial timestep/controller (in non DWR case), possibly dx_obs (see get_dx)
    def init_parameters(self, file_read, tol, DG = None, dwr_scheme = 0, diagnostics = 0, dt_cont_type = None,
                        time_int_class = 0, err_est_type = None):
        if dt_cont_type is None:
            dt_cont_type = file_read['dt_cont_type']
        if err_est_type is None:
            err_est_type = file_read['err_est_type']
        if DG is None:
            DG = file_read['dg']
        
        ## initialize problem class
        if file_read['problem_type'] == -1:
            from problem_verify import Problem_verification
            Prob = Problem_verification(DG = DG, t_end = file_read['problem_t_end'], func = file_read['func_type'],
                                        dwr_scheme = dwr_scheme)
        elif file_read['problem_type'] == -2:
            from problem_verify import Problem_verification_time
            Prob = Problem_verification_time(DG = DG, t_end = file_read['problem_t_end'], func = file_read['func_type'],
                                             dwr_scheme = dwr_scheme)
        elif file_read['problem_type'] == 0:
            from problem_basic    import Problem_Basic
            Prob = Problem_Basic(DG = DG, t_end = file_read['problem_t_end'], func = file_read['func_type'],
                                 para_stiff = file_read['problem_stiff_para'], dwr_scheme = dwr_scheme)
        elif file_read['problem_type'] == 12:
            from problem_src_adv import Problem_heat_source_adv
            Prob = Problem_heat_source_adv(func = file_read['func_type'], gridsize = file_read['problem_meshsize'],
                                           domain = file_read['problem_domain'], err_est_type = err_est_type,
                                           dwr_scheme = dwr_scheme, t_end = file_read['problem_t_end'],
                                           direction = file_read['direction'])
        elif file_read['problem_type'] == 20:
            if file_read['dg'] == 1:
                file_read['dg'] = 0
            from problem_FSI import Problem_FSI_1D
            Prob = Problem_FSI_1D(t_end = file_read['problem_t_end'], gridsize = file_read['problem_meshsize'],
                                  D1 = file_read['fsi_d1'], D2 = file_read['fsi_d2'], 
                                  K1 = file_read['fsi_k1'], K2 = file_read['fsi_k2'],
                                  flux = file_read['fsi_flux'], DG = file_read['dg'],
                                  u0 = file_read['fsi_u0'])
        elif file_read['problem_type'] == 21:
            if file_read['dg'] == 1:
                file_read['dg'] = 0
            from problem_FSI import Problem_FSI_2D
            Prob = Problem_FSI_2D(t_end = file_read['problem_t_end'], gridsize = file_read['problem_meshsize'],
                                  D1 = file_read['fsi_d1'], D2 = file_read['fsi_d2'], 
                                  K1 = file_read['fsi_k1'], K2 = file_read['fsi_k2'],
                                  flux = file_read['fsi_flux'], DG = file_read['dg'],
                                  u0 = file_read['fsi_u0'])
        elif file_read['problem_type'] == 30:
            from problem_3_comp import Problem_3_comp
            Prob = Problem_3_comp(DG = DG, t_end = file_read['problem_t_end'], func = file_read['func_type'],
                                  dwr_scheme = dwr_scheme)
        else:
            raise RuntimeError('invalid problem type')
        
        # further initialize problem class
        Prob.initialize_vectors(precond = file_read['precond'])
        
        # initialize time integration class
        if dwr_scheme == 0: # Non DWR
            if DG in [0, 3]: # RK case
                if file_read['problem_type'] == 20: # FSI case, needing special type of RK scheme
                    from time_int_FSI import SDIRK2_FSI_1D
                    osm = SDIRK2_FSI_1D(Prob.compute_matrices, t_end = Prob.t_end, tol_fixpoint = file_read['fsi_tol_fixpoint'])
                elif file_read['problem_type'] == 21: # FSI case, needing special type of RK scheme
                    from time_int_FSI import SDIRK2_FSI_2D
                    osm = SDIRK2_FSI_2D(Prob.compute_matrices, t_end = Prob.t_end, tol_fixpoint = file_read['fsi_tol_fixpoint'])
                else: # classic RK stuff
                    from time_int_RK import RK_get_class
                    osm = RK_get_class(rhs = Prob.get_rhs(), method_no = file_read['time_int_type'],
                                       t_end = Prob.t_end)
                Prob.do_step = osm.do_step
                order_low = osm.order_low
            else:
                order_low = Prob.order_low
                
            get_est_func = Prob.get_j_err_func(err_est_type)
                
            # initialize timestep controller
            from dt_controllers import Timestep_controller
            dt_cont = Timestep_controller(tol = tol, t_end = file_read['problem_t_end'], get_err = get_est_func,
                                          order_low = order_low, f_min = file_read['dt_cont_f_min'], f_max = file_read['dt_cont_f_max'],
                                          f_safety = file_read['dt_cont_f_safety'], get_y0 = Prob.get_initial_value, 
                                          get_eval_err = Prob.get_eval_err, j_val_func = Prob.j_val_func)
            # initialize time integration class
            if time_int_class == 0: # normal time integration class
                from time_int import get_Time_int
                Time_integ = get_Time_int(t_end = file_read['problem_t_end'], dg = file_read['dg'], diagnostics = diagnostics,
                                          get_eval = Prob.get_eval, get_eval_err = Prob.get_eval_err,
                                          get_sol_int = Prob.solution_int, j_val_func = Prob.j_val_func,
                                          get_j_err_func = Prob.get_j_err_func, j = Prob.j, get_sol_point = Prob.solution_point,
                                          get_y0 = Prob.get_initial_value)
            elif time_int_class == 1: # run_global_error case
                from run_global_error import Time_int_global_error
                Time_integ = Time_int_global_error(t_end = file_read['problem_t_end'], get_eval = Prob.get_eval,
                                                   get_eval_err = Prob.get_eval_err, get_y0 = Prob.get_initial_value)
            # pass dt-controller function to time integration class
            Time_integ.get_new_dt = dt_cont.get_controller(dt_cont_type)
            
        else: # DWR case
            if file_read['adjoint_steps'] is not None:
                file_read['dwr_start_cells'] = file_read['adjoint_steps']
            from time_int_DWR import Time_int_DWR
            Time_integ = Time_int_DWR(file_read['problem_t_end'], file_read['dwr_start_cells'],
                                      file_read['dwr_refine_perc'], tol = tol, get_eval = Prob.get_eval_DWR,
                                      get_sol_int = Prob.solution_int, j_val = Prob.j_val_func,
                                      get_dwr_est = Prob.get_dwr_est, get_y0 = Prob.get_initial_value)
                                      
        # return parameters/classes accordingly
        if dwr_scheme != 0:
            return Time_integ
        else:
            dt0_scheme = dt_cont.get_dt0_controller(file_read['dt_cont_dt0'])
            return Time_integ, dt0_scheme(dt_cont_type)
    
    ## General function for priting plots
    # @param self pointer to self
    # @param file_read parameters read from file
    # @param dpi dpi value for image saving
    # @return None
    def print_plot(self, file_read, dpi = 1000):
        pass
        