# -*- coding: utf-8 -*-
"""
Created on Wed Jun 22 17:08:36 2016
@author: Peter Meisrimel
"""
from __future__ import division
from dolfin import *
import numpy as np
import pylab as pl

## function to get Time integration class for either a DG or RK based problem
# @param t_end end time of problem
# @param diagnostics set to True to store more information during a forward run
# @param dg indicator on wether to use DG or RK, 1 for DG
# @param get_eval function for basic forward step evaluation
# @param get_eval_err function for forward step evaluation plus error estimate
# @param get_sol_int function to get reference solution of integral
# @param j_val_func function to access accumulated value of J
# @param get_j_err_func function to get error estimate for denominator of dt controller
# @param j density function
# @param get_sol_point function to get reference pointwise solution
# @param get_y0 function to get initial values of problem
# @return according Time integration class
def get_Time_int(t_end, diagnostics = 0, dg = 1, get_eval = None, get_eval_err = None, get_sol_int = None, j_val_func = None,
                 get_j_err_func = None, j = None, get_sol_point = None, get_y0 = None):
    if dg in [0, 3]:
        from time_int_RK import Time_int_RK
        return Time_int_RK(t_end = t_end, DG = 0, diagnostics = diagnostics, get_eval = get_eval,
                           get_eval_err = get_eval_err, get_sol_int = get_sol_int, j_val_func = j_val_func,
                           get_j_err_func = get_j_err_func, j = j, get_sol_point = get_sol_point, get_y0 = get_y0)
    elif dg in [1, 2]:
        return Time_int(t_end = t_end, DG = 1, diagnostics = diagnostics, get_eval = get_eval,
                        get_eval_err = get_eval_err, get_sol_int = get_sol_int, j_val_func = j_val_func,
                        get_j_err_func = get_j_err_func, j = j, get_sol_point = get_sol_point, get_y0 = get_y0)
    else:
        raise TypeError('invalid time integration scheme selected')

## DG based time-integration class for building and solving adjoint equations
# Contains methods to solve forward and backward in time, including adaptivity
class Time_int():
    ## time discretization
    times = []
    ## order of main scheme, Crank-Nicolson
    order = 2
    ## order of embedded scheme, Implicit Euler
    order_low = 1
    ## number of rhs evaluations done per step
    rhs_step = 3
    
    ## Initialize function
    # @param self pointer to itself
    # @param t_end end time of problem
    # @param DG indicator for DG (1) or RK (0)
    # @param diagnostics set to True to store more information during a forward run
    # @param get_eval function for basic forward step evaluation
    # @param get_eval_err function for forward step evaluation plus error estimate
    # @param get_sol_int function to get reference solution of integral
    # @param j_val_func function to access accumulated value of J
    # @param get_j_err_func function to get error estimate for denominator of dt controller
    # @param j density function
    # @param get_sol_point function to get reference pointwise solution
    # @param get_y0 function to get initial values of problem
    # @return None
    def __init__(self, t_end, DG = 0, diagnostics = 0, get_eval = None, get_eval_err = None, get_sol_int = None, j_val_func = None,
                 get_j_err_func = None, j = None, get_sol_point = None, get_y0 = None):
        ## end time
        self.t_end = t_end
        ## if DG or RK
        self.DG = DG
        ## will store more information about forward run if set to True
        self.diagnostics = diagnostics
        ## basic timestep evaluation function
        self.get_eval = get_eval
        ## timestep evaluation function plus local error estimate
        self.get_eval_err = get_eval_err
        ## get reference solution of integral value
        self.get_sol_int = get_sol_int
        ## function to get and set cummulative j value
        self.j_val_func = j_val_func
        ## function to get error estimation function
        self.get_j_err_func = get_j_err_func
        ## density function of quantity of interest
        self.j = j
        ## get reference solution for a given point in time (also local problems)
        self.get_sol_point = get_sol_point
        ## function to get initial value of problem
        self.get_y0 = get_y0
        
        if self.diagnostics == 1:
            ## diagnostic stats
            self.stats = {'t': [0]}
    
    ## Solve problem forward in time, given fixed timesteps
    # @param self pointer to itself
    # @param y_old current y value
    # @return final y(t_end)
    def forward(self, y_old):
        for i in range(len(self.times) - 1):
            y_old = self.get_eval(y_old, self.times[i], self.times[i+1] - self.times[i]) # make timestep
        return y_old
            
    
    ## Solve problem forward in time with variable step sizes
    # @param self pointer to itself
    # @param dt_val size of initial timestep
    # @param y_old current y value
    # @return None, relevant j value is stored in problem class and accessed via get_j_func
    def forward_le(self, dt_val, y_old):
        # initialize
        prev_time = self.times[0]
        while prev_time < self.t_end:
            if dt_val < 1.e-15:
                raise ValueError('timestep critically low with dt = {} at t = {}'.format(dt_val, prev_time))
            # get new time
            next_time = min([self.times[-1] + dt_val, self.t_end])
            dt_val = next_time - prev_time
            self.times.append(next_time)
            
            # get error estimate and new timestep
            y_new, l_err = self.get_eval_err(y_old, prev_time, dt_val)
            if self.diagnostics:
                self.get_diagnostics(dt_val)
            dt_val = self.get_new_dt(y_new, y_old, l_err, dt_val)
            prev_time = next_time
            y_old = y_new
    
    ## function to extract extra information from a forward run \n
    # can be extended to include more variables one wishes to monitor
    # @param self pointer to itself
    # @param dt_val current timestep
    # @return None, values are written into self.stats
    def get_diagnostics(self, dt_val):
        if self.diagnostics >= 1:
            dt_val = np.float64(dt_val)
            self.stats['t'].append(self.stats['t'][-1] + dt_val)
    
    ## Run with DG discretization
    # @param self pointer to itself
    # @param dt0 initial timestep, default: t_end
    # @param adaptive set to 0 for constant timesteps, 1 for adaptivity
    # @param final Set to True for returning y(t_end) instead of normal return
    # @return Error, runtime, number of steps done (unless final == True)
    def run(self, dt0 = None, adaptive = 1, final = False):
        if dt0 == None: dt0 = self.t_end
        import time
        start_time = time.time()
        
        # do normal forward run
        if adaptive == 0: # fixed dt
            self.times = list(np.arange(0, self.t_end, dt0)) + [self.t_end]
            y_final = self.forward(self.get_y0())
            if final:
                return y_final
        elif adaptive == 1: # time adaptive
            self.times = [0]
            self.forward_le(dt0, self.get_y0())
        else:
            raise RuntimeError('invalid adaptive parameter')
        forward_time = time.time() - start_time
        
        # get error by reference solution
        true_error = float(abs(self.get_sol_int() - self.j_val_func()))
        steps = len(self.times) - 1
        print('{} steps done in {} seconds'.format(steps, forward_time))
        return true_error, forward_time, steps