# -*- coding: utf-8 -*-
"""
Created on Fri Feb 02 09:54:36 2017
@author: Peter Meisrimel
credit to Patrick Farell
"""
from __future__ import division
from dolfin import *
import numpy as np

## DG based time-integration class for building and solving adjoint equations
# Contains methods to solve forward and backward in time, including adaptivity
class Time_int_DWR():
    ## forward solution
    solns = {}
    ## backward solution
    adjsolns = {}
    ## Initialize function
    # @param self pointer to itself
    # @param t_end end time
    # @param start_cells number of cells in initial grid
    # @param perc percentage for refinement
    # @param tol target tolerance
    # @param get_eval function for basic forward step evaluation
    # @param get_sol_int function to obtain reference solution for error calculation
    # @param j_val function for getting and setting J value which is accumulated during computation, see problem.py for details
    # @param get_dwr_est function for dwr estimate calculation
    # @param get_y0 function to get initial values from problem, see problem.py for details
    # @return None
    def __init__(self, t_end, start_cells, perc, tol, get_eval = None, get_sol_int = None,
                 j_val = None, get_dwr_est = None, get_y0 = None):
        ## time discretization
        self.mesh = IntervalMesh(int(start_cells), 0, int(t_end)) 
        ## percentage of cells to refine in a single refinement step
        self.perc = perc
        ## target tolerance
        self.tol = tol
        ## end time
        self.t_end = t_end
        ## timestep evaluation function
        self.get_eval_DWR = get_eval
        ## function to get reference solution of integral
        self.get_sol_int = get_sol_int
        ## function to get and manipulate j_value accumulated
        self.j_val = j_val 
        ## dwr estimate function
        self.get_dwr_est = get_dwr_est
        ## function to pointer to y_old, needed here for annotation purposes, call using DWR = True
        self.get_y0 = get_y0
    
    ## Solve problem forward in time, given fixed timesteps
    # @param self pointer to itself
    # @param mesh mesh used for solving
    # @return None, forward solution is class parameter
    def forward(self, mesh):
        self.y_old = self.get_y0(DWR = True)
        self.j_val(0)
        solns = {}
        solns[0] = Function(self.y_old)
        times = list(mesh.coordinates().ravel())
        times.sort()
        for i in range(len(times)-1):
            t, t_new = times[i], times[i+1]
            y_new = self.get_eval_DWR(self.y_old, t, t_new - t, fwd = True) # make timestep
            solns[t_new] = Function(y_new) # store new value
            self.y_old.assign(y_new)
        return solns
    
    ## Solving of adjoint equation by LU factorization
    # @param self pointer to itself
    # @param mesh mesh used for solving
    # @return None, adjoint soltuion is class parameter
    def solve_adjoint(self, mesh):
        self.z_old = self.get_y0(DWR = True, adj = True)
        adj_solns = {}
        adj_solns[self.t_end] = Function(self.z_old)
        times = list(mesh.coordinates().ravel())
        times.sort()
        times.reverse()
        for i in range(len(times)-1):
            t, t_new = times[i], times[i+1]
            z_new = self.get_eval_DWR(self.z_old, t, t_new - t, fwd = False) # make timestep
            adj_solns[t_new] = Function(z_new) # store new value
            self.z_old.assign(z_new)
        return adj_solns
    
    ## Run with DG discretization, might include DWR estimation
    # @param self pointer to itself
    # @param mesh mesh from previous tolerance, None in case of initial mesh
    # @param est estimate from previous tolerance, None in case of initial mesh
    # @param est_list error estimate list from previous tolerance, None in case of initial mesh
    # @param j_prev functional result from previous run
    # @return Error, DWR estimate, runtime of forward run, runtime of DWR estimation
    def run(self, mesh = None, est = np.inf, est_list = None, j_prev = 0):
        # runtime measurements
        import time
        start_time = time.time()
        if mesh is not None:
            self.mesh = mesh
        while True:
            if est <= self.tol:
                if self.j_val() == 0: # previous results still sufficient
                    self.j_val(j_prev)
                break
            elif est_list is not None: # True in for the initial grid
                self.mesh = self.refine_basic(self.mesh, self.perc, abs(est_list))
            sol_coarse = self.forward(self.mesh)
            adj_coarse = self.solve_adjoint(self.mesh)
            mesh_fine = self.refine_basic(self.mesh, 1, [1]*(len(sol_coarse.keys()) - 1))
            
            self.forward(mesh_fine)
            adj_fine = self.solve_adjoint(mesh_fine)
            
            t_fine = adj_fine.keys()
            t_fine.sort()
            
            est_list = self.get_dwr_est(sol_coarse, adj_coarse, adj_fine)
            est = sum(est_list)
            print('DWR est : {} for {} cells'.format(est, self.mesh.num_cells()))
        true_error = abs(self.get_sol_int() - self.j_val())
        print('TRUE ERROR', true_error)
        dwr_time = time.time() - start_time
        
        return est, dwr_time, true_error, est_list, self.j_val(), self.mesh.num_cells()
    
    ## DWR refinement routine \n
    # splits 'self.perc' of all timeinterval into half, starting by the ones with largest weights
    # @param self pointer to self
    # @param mesh mesh to refine
    # @param perc percentage of mesh cells to refine
    # @param weights DWR error weights 
    # @return (list) new and refine time discretization
    def refine_basic(self, mesh, perc, weights):
        num = int(np.ceil(len(weights)*perc))
        markers = CellFunction('bool', mesh)
        comb = zip(weights, range(len(weights)))
        comb.sort(key = lambda a : a[0], reverse = True) # sort by weights
        for i in range(num):
            markers[comb[i][1]] = True
        return refine(mesh, markers)
    
    ## simple routine to solve forward and backward problem, returning solution of backward problem
    # @param self pointer to self
    # @return list of times used, dictionary of adjoint solution with times as keys
    def run_adj(self):
        self.forward(self.mesh)
        adj_sol = self.solve_adjoint(self.mesh)
        times = list(self.mesh.coordinates().ravel())
        times.sort()
        return times, adj_sol