# -*- coding: utf-8 -*-
"""
Created on Fri Aug 12 13:39:21 2016
@author: Peter Meisrimel
"""
from __future__ import division
from scipy.optimize import newton_krylov
from time_int import Time_int
import numpy as np

## RK time integration class, general time integration procedures
class Time_int_RK(Time_int):
    ## Forward solve given fixed points in time
    # @param self pointer to itself
    # @param y_old current y value
    # @return None
    def forward(self, y_old):
        for i in range(len(self.times) - 1):
            y_old = self.get_eval(y_old, self.times[i], self.times[i+1] - self.times[i]) # make timestep
        return y_old
    
    ## Forward solve with variable step sizes
    # @param self pointer to itself
    # @param dt_val initial timestep
    # @param y_old current y value
    # @return None
    def forward_le(self, dt_val, y_old):
        prev_time = self.times[0]
        while prev_time < self.t_end:
            # get new time
            next_time = min([self.times[-1] + dt_val, self.t_end])
            dt_val = next_time - prev_time
            self.times.append(next_time)
            
            # do step, get error estimate and new timestep
            y_new, l_err = self.get_eval_err(y_old, prev_time, dt_val)
            if self.diagnostics:
                self.get_diagnostics(dt_val)
            dt_val = self.get_new_dt(y_new, y_old, l_err, dt_val)
            prev_time = next_time
            y_old = y_new
        
## Parent class for general one step methods, embedded RK schemes
class One_step_method():
    ## first same as last property \todo is not formally implemented, only used in calculation of number of timesteps done
    fsal = False
    ## Basic initialization
    # @param self pointer to itself
    # @param rhs right-hand side function
    # @param t_end end time of simulation
    # @return None
    def __init__(self, rhs, t_end):
        ## right-hand side function
        self.rhs = rhs
        ## c vector of butcher array
        self.c = np.zeros(self.stages) # stages & A to be set in child class
        for i in range(self.stages):
            for j in range(self.stages):
                self.c[i] = self.c[i] + self.A[i,j]
        ## placeholder list for stage derivatives
        self.k = list(range(self.stages))
        ## final time
        self.t_end = t_end

## Contruct root problem for DIRK schemes to be used in Newton
# @param t current time point
# @param dt timestep
# @param y y vector
# @param rhs right-hand side function
# @param gamma diagonal value of SDIRK scheme
# @param c c value corresponding to current stage
# @return pointer to root problem function
def f_root(t, dt, y, rhs, gamma = 1, c = 1):
    def f(x):
        return rhs(t + c*dt, y + dt*gamma*x) - x
    return f

## Function to select and initialize correct RK scheme according to input file 
# @param rhs right-hand side function
# @param method_no indentifier for method to be used
# @param t_end end time of problem
# @return pointer to initialized scheme
def RK_get_class(rhs, method_no, t_end):
    if method_no == -21:   return Impl_SDIRK_21(rhs, t_end)
    elif method_no == -32: return Impl_SDIRK_32(rhs, t_end)
    elif method_no == -31: return Impl_SDIRK_31(rhs, t_end)
    elif method_no == 21:  return Expl_RK_21   (rhs, t_end)
    elif method_no == 32:  return Expl_RK_32   (rhs, t_end)
    elif method_no == 31:  return Expl_RK_31   (rhs, t_end)
    elif method_no == 45:  return Expl_RK_45   (rhs, t_end)
    elif method_no == 541: return Expl_RK_541  (rhs, t_end)
    elif method_no == 542: return Expl_RK_542  (rhs, t_end)
    elif method_no == 54:  return Expl_RK_542  (rhs, t_end)
    elif method_no == 53:  return Expl_RK_53   (rhs, t_end)
    elif method_no == 52:  return Expl_RK_52   (rhs, t_end)
    elif method_no == 51:  return Expl_RK_51   (rhs, t_end)
    raise ValueError('Scheme with desired orders not (yet) implemented')

## Generic DRIK scheme
class Impl_DIRK(One_step_method):
    ## general function to perform a single step
    # @param self pointer to itself
    # @param y_old current y value
    # @param t current time
    # @param dt_val size of timestep
    # @return new y value and lower order approximation (embedded solution)
    def do_step(self, y_old, t, dt_val):
        dt = min(dt_val, self.t_end - t)
        y = np.copy(y_old)

        for i in range(self.stages):
            temp = np.copy(y_old)
            for j in range(i):
                temp = temp + dt * self.A[i,j] * self.k[j]
            try:
                self.k[i] = newton_krylov(f_root(t, dt, temp, self.rhs, self.A[i,i], self.c[i]),
                                          y if i == 0 else self.k[i-1], f_tol = 1.e-10, f_rtol = 1.e-10)
            except ValueError:
                self.k[i] = newton_krylov(f_root(t, dt, temp, self.rhs, self.A[i,i], self.c[i]),
                                          y, f_tol = 1.e-10, f_rtol = 1.e-10)
        
        ynew = np.copy(y_old)
        yaux = np.copy(y_old)
        for i in range(self.stages):
            ynew = ynew + dt * self.b[i]     * self.k[i]
            yaux = yaux + dt * self.b_hat[i] * self.k[i]

        return ynew, ynew - yaux
        
## Generic explicit RK method
class Expl_RK(One_step_method):
    ## general function to perform a single step
    # @param self pointer to itself
    # @param y_old current y value
    # @param t current time
    # @param dt_val size of timestep
    # @return new y value and lower order approximation (embedded solution)
    def do_step(self, y_old, t, dt_val):
        dt = min(dt_val, self.t_end - t)

        self.k[0] = self.rhs(t, y_old)
        for i in range(1, self.stages):
            temp = np.copy(y_old)
            for j in range(i):
                temp = temp + dt * self.A[i,j] * self.k[j]
            self.k[i] = self.rhs(t + self.c[i]*dt, temp)
        
        ynew = np.copy(y_old)
        yaux = np.copy(y_old)
        for i in range(self.stages):
            ynew = ynew + dt * self.b[i]     * self.k[i]
            yaux = yaux + dt * self.b_hat[i] * self.k[i]

        return ynew, ynew - yaux

## Implementation of an explicit scheme of orders (2,1)
# \n Heun-Euler method
class Expl_RK_21(Expl_RK):
    ## orders of method
    order, order_low = 2, 1
    ## stages of scheme
    stages = 2
    ## A matrix of butcher array
    A = np.array([[0, 0],[1, 0]])
    ## coefficient vector of main scheme
    b = np.array([1/2, 1/2])
    ## coefficient vector of embedded scheme
    b_hat = np.array([1, 0])

## Implementation of an explicit scheme of orders (3,2)
# \n Bogacki-Shampine method
class Expl_RK_32(Expl_RK):
    ## orders of method
    order, order_low = 3, 2
    ## stages of scheme
    stages = 4
    ## A matrix of butcher array
    A = np.array([[0, 0, 0, 0],[1/2, 0, 0, 0],[0, 3/4, 0, 0],[2/9, 1/3, 4/9, 0]])
    ## coefficient vector of main scheme
    b = np.array([2/9, 1/3, 4/9, 0])
    ## coefficient vector of embedded scheme
    b_hat = np.array([7/24, 1/4, 1/3, 1/8])
    ## first same as last property
    fsal = True

## Implementation of an explicit scheme of orders (3,1)
# \n Bogacki-Shampine base with explicit euler as estimator, this one is very inefficient, testing purposes only
class Expl_RK_31(Expl_RK_32):
    ## order of embedded method
    order_low = 1
    ## coefficient vector of embedded scheme
    b_hat = np.array([1., 0, 0, 0])
        
## Implementation of an explicit scheme of orders (4,5)
# \n Runge-Kutta Fehlberg method
class Expl_RK_45(Expl_RK):
    ## orders of method
    order, order_low = 5, 4
    ## stages of scheme
    stages = 6
    ## A matrix of butcher array
    A = np.array([[0., 0., 0., 0., 0., 0.],
                  [1/4, 0., 0., 0., 0., 0.],
                  [3/32, 9/32, 0., 0., 0., 0.],
                  [1932/2197, -7200/2197, 7296/2197, 0., 0., 0.],
                  [439/216, -8., 3680/513, -845/4104, 0., 0.],
                  [-8/27, 2., -3544/2565, 1859/4104, -11/40, 0.]])
    ## coefficient vector of main scheme
    b = np.array([25/216, 0., 1408/2565, 2197/4104, -1/5, 0.])
    ## coefficient vector of embedded scheme
    b_hat = np.array([16/135, 0., 6656/12825, 28561/56430, -9/50, 2/55])
        
## Implementation of an explicit scheme of orders (5,4)
# \n Cash-Karp method
class Expl_RK_541(Expl_RK):
    ## orders of method
    order, order_low = 5, 4
    ## stages of scheme
    stages = 6
    ## A matrix of butcher array
    A = np.array([[0., 0., 0., 0., 0., 0.],
                  [1/5, 0., 0., 0., 0., 0.],
                  [3/40, 9/40, 0., 0., 0., 0.],
                  [3/10, -9/10, 6/5, 0., 0., 0.],
                  [-11/54, 5/2, -70/27, 35/27, 0., 0.],
                  [1631/55296, 175/512, 575/13824, 44275/110592, 253/4096, 0.]])
    ## coefficient vector of main scheme
    b = np.array([37/378, 0., 250/621, 125/594, 0., 512/1771])
    ## coefficient vector of embedded scheme
    b_hat = np.array([2825/27648, 0., 18575/48384, 13525/55296, 277/14336, 1/4])

## Implementation of an explicit scheme of orders (5,3)
# \n Cash-Karp method base with embedded 3rd order scheme
class Expl_RK_53(Expl_RK_541):
    ## order of embedded scheme
    order_low = 3
    ## coefficient vector of embedded scheme
    b_hat = np.array([2/9, 0., 0., 25/36, 1/12, 0.])

## Implementation of an explicit scheme of orders (5,2)
# \n Cash-Karp method base with embedded 2nd order scheme
class Expl_RK_52(Expl_RK_541):
    ## order of embedded scheme
    order_low = 2
    ## coefficient vector of embedded scheme
    b_hat = np.array([-2/3, 0., 5/3, 0., 0., 0.])

## Implementation of an explicit scheme of orders (5,1)
# \n Cash-Karp method base with embedded 1st order scheme
class Expl_RK_51(Expl_RK_541):
    ## order of embedded scheme
    order_low = 1
    ## coefficient vector of embedded scheme
    b_hat = np.array([1., 0., 0., 0., 0., 0.])
        
## Implementation of an explicit scheme of orders (5,4)
# \n Dormand-Prince method
class Expl_RK_542(Expl_RK):
    ## orders of method
    order, order_low = 5, 4
    ## stages of scheme
    stages = 7
    ## A matrix of butcher array
    A = np.array([[0., 0., 0., 0., 0., 0., 0.],
                  [1/5, 0., 0., 0., 0., 0., 0.],
                  [3/40, 9/40, 0., 0., 0., 0., 0.],
                  [44/45, -56/15, 32/9, 0., 0., 0., 0.],
                  [19372/6561, -25360/2187, 64448/6561, -212/729, 0., 0., 0.],
                  [9017/3168, -355/33, 46732/5247, 49/176, -5103/18656, 0., 0.],
                  [35/384, 0., 500/1113, 125/192, -2187/6784, 11/84, 0.]])
    ## coefficient vector of main scheme
    b = np.array([35/384, 0., 500/1113, 125/192, -2187/6784, 11/84, 0.])
    ## coefficient vector of embedded scheme
    b_hat = np.array([5179/57600, 0., 7571/16695, 393/640, -92097/339200, 187/2100, 1/40])
    ## first same as last property
    fsal = True
        
## Implementation of an DIRK scheme of orders (2,1)
# \n Method of Ellsiepen
class Impl_SDIRK_21(Impl_DIRK):
    ## orders of method
    order, order_low = 2, 1
    ## stages of scheme
    stages = 2
    alpha = 1 - 1/np.sqrt(2)
    alpha_hat = 2 - 5/4*np.sqrt(2)
    ## A matrix of butcher array
    A = np.array([[alpha, 0],[1 - alpha, alpha]])
    ## coefficient vector of main scheme
    b = np.array([1 - alpha, alpha])
    ## coefficient vector of embedded scheme
    b_hat = np.array([1 - alpha_hat, alpha_hat])

## Implementation of an DIRK scheme of orders (3,2)
# \n Method of Cash
class Impl_SDIRK_32(Impl_DIRK):
    ## orders of method
    order, order_low = 3, 2
    ## stages of scheme
    stages = 3
    
    alpha = 1.2084966491760101
    beta = -0.6443631706844691
    gamma = 0.4358665215084580
    tau_minus = 0.2820667392457705
    alpha_hat = 0.7726301276675511
    beta_hat = 0.2273698723324489
    
    ## A matrix of butcher array
    A = np.array([[gamma, 0, 0], [tau_minus, gamma, 0], [alpha, beta, gamma]])
    ## coefficient vector of main scheme
    b = np.array([alpha, beta, gamma])
    ## coefficient vector of embedded scheme
    b_hat = np.array([alpha_hat, beta_hat, 0])

## Implementation of an DIRK scheme of orders (3,2)
# \n Method of Cash base with embedded implicit Euler scheme       
class Impl_SDIRK_31(Impl_SDIRK_32):
    ## order of embedded scheme
    order_low = 1
    ## coefficient vector of embedded scheme
    b_hat = np.array([1, 0, 0])